import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.modeling.reversible_embedding import (
    ReversibleEmbedding,
)
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.gemma.rms_normalization import RMSNormalization
from keras_hub.src.models.t5gemma.t5gemma_decoder import T5GemmaDecoderLayer
from keras_hub.src.models.t5gemma.t5gemma_encoder import T5GemmaEncoderLayer
from keras_hub.src.models.t5gemma.t5gemma_layers import (
    t5gemma_kernel_initializer,
)
from keras_hub.src.utils.keras_utils import clone_initializer


@keras_hub_export("keras_hub.models.T5GemmaBackbone")
class T5GemmaBackbone(Backbone):
    """T5Gemma backbone model.

    This class implements the encoder-decoder backbone of the T5Gemma model,
    consisting of an embedding layer, a stack of encoder layers, and a
    stack of decoder layers.

    Args:
        vocabulary_size: int, The size of the vocabulary.
        encoder_hidden_dim: int, The hidden dimensionality of the encoder.
        encoder_intermediate_dim: int, The intermediate size of the encoder's
            feed-forward networks.
        encoder_num_layers: int, The number of encoder layers.
        encoder_num_attention_heads: int, The number of attention heads in the
            encoder.
        encoder_num_key_value_heads: int, The number of key-value heads in the
            encoder.
        encoder_head_dim: int, The dimensionality of each attention head in the
            encoder.
        encoder_layer_types: list of str, A list of strings specifying the type
            of attention layer for each encoder layer. Each element can be
            either `"sliding_attention"` or `"full_attention"`. For example,
            `["full_attention", "sliding_attention", ...]`.
        decoder_hidden_dim: int, The hidden dimensionality of the decoder.
        decoder_intermediate_dim: int, The intermediate size of the decoder's
            feed-forward networks.
        decoder_num_layers: int, The number of decoder layers.
        decoder_num_attention_heads: int, The number of attention heads in the
            decoder.
        decoder_num_key_value_heads: int, The number of key-value heads in the
            decoder.
        decoder_head_dim: int, The dimensionality of each attention head in the
            decoder.
        decoder_layer_types: list of str, A list of strings specifying the type
            of attention layer for each decoder layer. Each element can be
            either `"sliding_attention"` or `"full_attention"`. For example,
            `["full_attention", "sliding_attention", ...]`.
        dropout_rate: float, The dropout rate applied throughout the model.
            Defaults to `0.0`.
        rms_norm_eps: float, The epsilon value for RMS normalization. Defaults
            to `1e-6`.
        query_pre_attn_scalar: float, Scalar to multiply queries by before
            attention. Defaults to `1.0`.
        attention_bias: bool, Whether to include bias in attention computations.
            Defaults to `False`.
        hidden_activation: str, The activation function used in the feed-forward
            networks. Defaults to `"gelu_approximate"`.
        tie_word_embeddings: bool, Whether to tie input and output word
            embeddings. Defaults to `True`.
        initializer_range: float, The range for the random normal initializer.
            Defaults to `0.02`.
        attention_dropout: float, The dropout rate applied to attention weights.
            Defaults to `0.0`.
        sliding_window: int, optional, The window size for sliding attention.
            Required if any `layer_type` is `"sliding_attention"`. Defaults to
            `None`.
        cross_attention_hidden_size: int, optional, The hidden size for
            cross-attention in the decoder layers. If None, it defaults to
            `encoder_hidden_dim`. Defaults to `None`.
        attn_logit_softcapping: float, optional, The softcapping value for
            attention logits. Defaults to `None`.
        final_logit_softcapping: float, optional, The softcapping value for
            final logits. Defaults to `None`.
        rope_max_wavelength: float, The maximum wavelength for Rotary Positional
            Embeddings. Defaults to `10000.0`.
        dtype: string or `keras.mixed_precision.DTypePolicy`. The dtype to use
            for model computations and weights. Note that some computations,
            such as softmax and layer normalization, will always be done at
            float32 precision regardless of dtype. Defaults to `None`.
        **kwargs: Additional keyword arguments passed to the parent `Backbone`
            class.

    Examples:
    ```python
    import numpy as np
    from keras_hub.models import T5GemmaBackbone

    input_data = {
        "encoder_token_ids": np.ones(shape=(1, 12), dtype="int32"),
        "encoder_padding_mask": np.array(
            [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0]], dtype="int32"
        ),
        "decoder_token_ids": np.ones(shape=(1, 8), dtype="int32"),
        "decoder_padding_mask": np.array(
            [[1, 1, 1, 1, 1, 1, 1, 1]], dtype="int32"
        ),
    }

    # Randomly initialized T5Gemma backbone with custom config.
    model = T5GemmaBackbone(
        vocabulary_size=32000,
        # Encoder parameters.
        encoder_hidden_dim=256,
        encoder_intermediate_dim=512,
        encoder_num_layers=4,
        encoder_num_attention_heads=4,
        encoder_num_key_value_heads=2,
        encoder_head_dim=64,
        encoder_layer_types=["full_attention"] * 4,
        # Decoder parameters.
        decoder_hidden_dim=256,
        decoder_intermediate_dim=512,
        decoder_num_layers=4,
        decoder_num_attention_heads=4,
        decoder_num_key_value_heads=2,
        decoder_head_dim=64,
        decoder_layer_types=["full_attention"] * 4,
        # Common parameters.
        dropout_rate=0.1,
        rms_norm_eps=1e-6,
        query_pre_attn_scalar=1.0,
        attention_bias=False,
        hidden_activation="gelu_approximate",
    )
    output = model(input_data)
    ```
    """

    def __init__(
        self,
        vocabulary_size,
        encoder_hidden_dim,
        encoder_intermediate_dim,
        encoder_num_layers,
        encoder_num_attention_heads,
        encoder_num_key_value_heads,
        encoder_head_dim,
        encoder_layer_types,
        decoder_hidden_dim,
        decoder_intermediate_dim,
        decoder_num_layers,
        decoder_num_attention_heads,
        decoder_num_key_value_heads,
        decoder_head_dim,
        decoder_layer_types,
        dropout_rate=0.0,
        rms_norm_eps=1e-6,
        query_pre_attn_scalar=1.0,
        attention_bias=False,
        hidden_activation="gelu_approximate",
        tie_word_embeddings=True,
        initializer_range=0.02,
        attention_dropout=0.0,
        sliding_window=None,
        cross_attention_hidden_size=None,
        attn_logit_softcapping=None,
        final_logit_softcapping=None,
        rope_max_wavelength=10000.0,
        dtype=None,
        **kwargs,
    ):
        self.kernel_initializer = t5gemma_kernel_initializer(initializer_range)

        # === Layers ===
        self.token_embedding = keras.layers.Embedding(
            input_dim=vocabulary_size,
            output_dim=encoder_hidden_dim,
            embeddings_initializer=clone_initializer(self.kernel_initializer),
            dtype=dtype,
            name="encoder_token_embedding",
        )
        self.decoder_token_embedding = ReversibleEmbedding(
            input_dim=vocabulary_size,
            output_dim=decoder_hidden_dim,
            tie_weights=tie_word_embeddings,
            embeddings_initializer=clone_initializer(self.kernel_initializer),
            dtype=dtype,
            name="decoder_token_embedding",
        )
        self.encoder_layers = [
            T5GemmaEncoderLayer(
                hidden_size=encoder_hidden_dim,
                rms_norm_eps=rms_norm_eps,
                num_attention_heads=encoder_num_attention_heads,
                num_key_value_heads=encoder_num_key_value_heads,
                query_pre_attn_scalar=query_pre_attn_scalar,
                attention_bias=attention_bias,
                intermediate_size=encoder_intermediate_dim,
                hidden_activation=hidden_activation,
                head_dim=encoder_head_dim,
                dropout_rate=dropout_rate,
                initializer_range=initializer_range,
                attention_dropout=attention_dropout,
                layer_type=encoder_layer_types[i],
                sliding_window=sliding_window,
                attn_logit_softcapping=attn_logit_softcapping,
                rope_max_wavelength=rope_max_wavelength,
                name=f"encoder_layer_{i}",
                dtype=dtype,
            )
            for i in range(encoder_num_layers)
        ]
        self.encoder_norm = RMSNormalization(epsilon=rms_norm_eps, dtype=dtype)
        self.encoder_dropout = keras.layers.Dropout(dropout_rate, dtype=dtype)
        self.decoder_layers = [
            T5GemmaDecoderLayer(
                hidden_size=decoder_hidden_dim,
                rms_norm_eps=rms_norm_eps,
                num_attention_heads=decoder_num_attention_heads,
                num_key_value_heads=decoder_num_key_value_heads,
                query_pre_attn_scalar=query_pre_attn_scalar,
                attention_bias=attention_bias,
                intermediate_size=decoder_intermediate_dim,
                hidden_activation=hidden_activation,
                dropout_rate=dropout_rate,
                initializer_range=initializer_range,
                head_dim=decoder_head_dim,
                attention_dropout=attention_dropout,
                layer_type=decoder_layer_types[i],
                sliding_window=sliding_window,
                cross_attention_hidden_size=(
                    cross_attention_hidden_size or encoder_hidden_dim
                ),
                attn_logit_softcapping=attn_logit_softcapping,
                rope_max_wavelength=rope_max_wavelength,
                name=f"decoder_layer_{i}",
                dtype=dtype,
            )
            for i in range(decoder_num_layers)
        ]
        self.decoder_norm = RMSNormalization(epsilon=rms_norm_eps, dtype=dtype)
        self.decoder_dropout = keras.layers.Dropout(dropout_rate, dtype=dtype)

        # === Functional Model ===
        encoder_token_id_input = keras.Input(
            shape=(None,), dtype="int32", name="encoder_token_ids"
        )
        encoder_padding_mask_input = keras.Input(
            shape=(None,), dtype="int32", name="encoder_padding_mask"
        )
        decoder_token_id_input = keras.Input(
            shape=(None,), dtype="int32", name="decoder_token_ids"
        )
        decoder_padding_mask_input = keras.Input(
            shape=(None,), dtype="int32", name="decoder_padding_mask"
        )

        # Encoder.
        encoder_embeddings = self.token_embedding(encoder_token_id_input)
        encoder_embeddings = encoder_embeddings * keras.ops.cast(
            keras.ops.sqrt(encoder_hidden_dim), encoder_embeddings.dtype
        )
        encoder_hidden_states = self.encoder_dropout(encoder_embeddings)
        for layer in self.encoder_layers:
            encoder_hidden_states = layer(
                encoder_hidden_states, padding_mask=encoder_padding_mask_input
            )
        encoder_output = self.encoder_norm(encoder_hidden_states)
        encoder_output = self.encoder_dropout(encoder_output)

        # Decoder.
        decoder_embeddings = self.decoder_token_embedding(
            decoder_token_id_input
        )
        decoder_embeddings = decoder_embeddings * keras.ops.cast(
            keras.ops.sqrt(decoder_hidden_dim), decoder_embeddings.dtype
        )
        decoder_hidden_states = self.decoder_dropout(decoder_embeddings)
        for layer in self.decoder_layers:
            decoder_hidden_states, _ = layer(
                (decoder_hidden_states, encoder_output),
                self_attention_padding_mask=decoder_padding_mask_input,
                cross_attention_padding_mask=encoder_padding_mask_input,
            )
        decoder_output = self.decoder_norm(decoder_hidden_states)
        decoder_output = self.decoder_dropout(decoder_output)

        super().__init__(
            inputs={
                "encoder_token_ids": encoder_token_id_input,
                "encoder_padding_mask": encoder_padding_mask_input,
                "decoder_token_ids": decoder_token_id_input,
                "decoder_padding_mask": decoder_padding_mask_input,
            },
            outputs={
                "encoder_sequence_output": encoder_output,
                "decoder_sequence_output": decoder_output,
            },
            dtype=dtype,
            **kwargs,
        )

        # === Config ===
        self.encoder_hidden_dim = encoder_hidden_dim
        self.encoder_intermediate_dim = encoder_intermediate_dim
        self.encoder_num_layers = encoder_num_layers
        self.encoder_num_attention_heads = encoder_num_attention_heads
        self.encoder_num_key_value_heads = encoder_num_key_value_heads
        self.encoder_head_dim = encoder_head_dim
        self.encoder_layer_types = encoder_layer_types
        self.decoder_hidden_dim = decoder_hidden_dim
        self.decoder_intermediate_dim = decoder_intermediate_dim
        self.decoder_num_layers = decoder_num_layers
        self.decoder_num_attention_heads = decoder_num_attention_heads
        self.decoder_num_key_value_heads = decoder_num_key_value_heads
        self.decoder_head_dim = decoder_head_dim
        self.decoder_layer_types = decoder_layer_types
        self.vocabulary_size = vocabulary_size
        self.dropout_rate = dropout_rate
        self.rms_norm_eps = rms_norm_eps
        self.tie_word_embeddings = tie_word_embeddings
        self.query_pre_attn_scalar = query_pre_attn_scalar
        self.attention_bias = attention_bias
        self.hidden_activation = hidden_activation
        self.initializer_range = initializer_range
        self.attention_dropout = attention_dropout
        self.sliding_window = sliding_window
        self.cross_attention_hidden_size = (
            cross_attention_hidden_size or encoder_hidden_dim
        )
        self.attn_logit_softcapping = attn_logit_softcapping
        self.final_logit_softcapping = final_logit_softcapping
        self.rope_max_wavelength = rope_max_wavelength

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary_size": self.vocabulary_size,
                "encoder_hidden_dim": self.encoder_hidden_dim,
                "encoder_intermediate_dim": self.encoder_intermediate_dim,
                "encoder_num_layers": self.encoder_num_layers,
                "encoder_num_attention_heads": self.encoder_num_attention_heads,
                "encoder_num_key_value_heads": self.encoder_num_key_value_heads,
                "encoder_layer_types": self.encoder_layer_types,
                "encoder_head_dim": self.encoder_head_dim,
                "decoder_hidden_dim": self.decoder_hidden_dim,
                "decoder_intermediate_dim": self.decoder_intermediate_dim,
                "decoder_num_layers": self.decoder_num_layers,
                "decoder_num_attention_heads": self.decoder_num_attention_heads,
                "decoder_num_key_value_heads": self.decoder_num_key_value_heads,
                "decoder_layer_types": self.decoder_layer_types,
                "decoder_head_dim": self.decoder_head_dim,
                "dropout_rate": self.dropout_rate,
                "rms_norm_eps": self.rms_norm_eps,
                "tie_word_embeddings": self.tie_word_embeddings,
                "query_pre_attn_scalar": self.query_pre_attn_scalar,
                "attention_bias": self.attention_bias,
                "hidden_activation": self.hidden_activation,
                "initializer_range": self.initializer_range,
                "attention_dropout": self.attention_dropout,
                "sliding_window": self.sliding_window,
                "cross_attention_hidden_size": self.cross_attention_hidden_size,
                "attn_logit_softcapping": self.attn_logit_softcapping,
                "final_logit_softcapping": self.final_logit_softcapping,
                "rope_max_wavelength": self.rope_max_wavelength,
            }
        )
        return config



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm_preprocessor import CausalLMPreprocessor
from keras_hub.src.models.qwen3.qwen3_backbone import Qwen3Backbone
from keras_hub.src.models.qwen3.qwen3_tokenizer import Qwen3Tokenizer


@keras_hub_export("keras_hub.models.Qwen3CausalLMPreprocessor")
class Qwen3CausalLMPreprocessor(CausalLMPreprocessor):
    backbone_cls = Qwen3Backbone
    tokenizer_cls = Qwen3Tokenizer



# Metadata for loading pretrained model weights.
backbone_presets = {
    "dfine_nano_coco": {
        "metadata": {
            "description": (
                "D-FINE Nano model, the smallest variant in the family, "
                "pretrained on the COCO dataset. Ideal for applications "
                "where computational resources are limited."
            ),
            "params": 3788625,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_nano_coco/1",
    },
    "dfine_small_coco": {
        "metadata": {
            "description": (
                "D-FINE Small model pretrained on the COCO dataset. Offers a "
                "balance between performance and computational efficiency."
            ),
            "params": 10329321,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_small_coco/1",
    },
    "dfine_medium_coco": {
        "metadata": {
            "description": (
                "D-FINE Medium model pretrained on the COCO dataset. A solid "
                "baseline with strong performance for general-purpose "
                "object detection."
            ),
            "params": 19621160,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_medium_coco/1",
    },
    "dfine_large_coco": {
        "metadata": {
            "description": (
                "D-FINE Large model pretrained on the COCO dataset. Provides "
                "high accuracy and is suitable for more demanding tasks."
            ),
            "params": 31344064,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_large_coco/1",
    },
    "dfine_xlarge_coco": {
        "metadata": {
            "description": (
                "D-FINE X-Large model, the largest COCO-pretrained variant, "
                "designed for state-of-the-art performance where accuracy "
                "is the top priority."
            ),
            "params": 62834048,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_xlarge_coco/1",
    },
    "dfine_small_obj365": {
        "metadata": {
            "description": (
                "D-FINE Small model pretrained on the large-scale Objects365 "
                "dataset, enhancing its ability to recognize a wider "
                "variety of objects."
            ),
            "params": 10623329,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_small_obj365/1",
    },
    "dfine_medium_obj365": {
        "metadata": {
            "description": (
                "D-FINE Medium model pretrained on the Objects365 dataset. "
                "Benefits from a larger and more diverse pretraining corpus."
            ),
            "params": 19988670,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_medium_obj365/1",
    },
    "dfine_large_obj365": {
        "metadata": {
            "description": (
                "D-FINE Large model pretrained on the Objects365 dataset for "
                "improved generalization and performance on diverse object "
                "categories."
            ),
            "params": 31858578,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_large_obj365/1",
    },
    "dfine_xlarge_obj365": {
        "metadata": {
            "description": (
                "D-FINE X-Large model pretrained on the Objects365 dataset, "
                "offering maximum performance by leveraging a vast number "
                "of object categories during pretraining."
            ),
            "params": 63348562,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_xlarge_obj365/1",
    },
    "dfine_small_obj2coco": {
        "metadata": {
            "description": (
                "D-FINE Small model first pretrained on Objects365 and then "
                "fine-tuned on COCO, combining broad feature learning with "
                "benchmark-specific adaptation."
            ),
            "params": 10329321,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_small_obj2coco/1",
    },
    "dfine_medium_obj2coco": {
        "metadata": {
            "description": (
                "D-FINE Medium model using a two-stage training process: "
                "pretraining on Objects365 followed by fine-tuning on COCO."
            ),
            "params": 19621160,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_medium_obj2coco/1",
    },
    "dfine_large_obj2coco_e25": {
        "metadata": {
            "description": (
                "D-FINE Large model pretrained on Objects365 and then "
                "fine-tuned on COCO for 25 epochs. A high-performance model "
                "with specialized tuning."
            ),
            "params": 31344064,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_large_obj2coco_e25/1",
    },
    "dfine_xlarge_obj2coco": {
        "metadata": {
            "description": (
                "D-FINE X-Large model, pretrained on Objects365 and fine-tuned "
                "on COCO, representing the most powerful model in this "
                "series for COCO-style tasks."
            ),
            "params": 62834048,
            "path": "d_fine",
        },
        "kaggle_handle": "kaggle://keras/d-fine/keras/dfine_xlarge_obj2coco/1",
    },
}



from keras_hub.src.models.parseq.parseq_backbone import PARSeqBackbone
from keras_hub.src.models.parseq.parseq_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, PARSeqBackbone)



import keras

from keras_hub.src.layers.modeling.transformer_decoder import TransformerDecoder
from keras_hub.src.layers.modeling.transformer_layer_utils import (
    compute_causal_mask,
)
from keras_hub.src.layers.modeling.transformer_layer_utils import (
    merge_padding_and_attention_mask,
)
from keras_hub.src.models.moonshine.moonshine_layers import MoonshineMLP
from keras_hub.src.models.moonshine.moonshine_layers import (
    moonshine_kernel_initializer,
)
from keras_hub.src.models.moonshine.moonshine_multi_head_attention import (
    MoonshineMultiHeadAttention,
)
from keras_hub.src.utils.keras_utils import clone_initializer


@keras.saving.register_keras_serializable(package="keras_hub")
class MoonshineDecoderBlock(TransformerDecoder):
    """Moonshine decoder block for sequence processing.

    This layer implements a decoder block that includes self-attention with
    causal masking, cross-attention with precomputed key/value pairs, and a
    feedforward network.

    Args:
        hidden_dim: int. The dimensionality of the model's hidden
            representations.
        intermediate_dim: int. The dimensionality of the intermediate
            representations in the feedforward network.
        num_heads: int. The number of attention heads for multi-head attention
            mechanisms.
        feedforward_expansion_factor: int, optional. A multiplicative factor for
            scaling the feedforward network dimension. Defaults to 4.
        use_swiglu_activation: bool, optional. Whether to use the SwiGLU
            activation in the feedforward network for improved performance.
            Defaults to True.
        pad_head_dim_to_multiple_of: int, optional. If specified, pads the head
            dimension to be a multiple of this value for performance
            optimization. Defaults to None.
        initializer_range: float, optional. The standard deviation of the
            truncated normal distribution used to initialize model weights.
            Defaults to 0.02.
        attention_bias: bool, optional. Whether to add a bias term to the
            attention computations. Defaults to False.
        attention_dropout: float, optional. The dropout rate applied to
            attention weights during training. Defaults to 0.0.
        dtype: str, optional. The data type to use for model computations and
            weights. Defaults to None.
        **kwargs: Additional keyword arguments passed to the base layer.
    """

    # References:
    # Defined and formulated based on the UsefulSensors implementation of the
    # DecoderLayer class (https://github.com/usefulsensors/moonshine/blob/4a000427bd36a1c2c6d20a86c672dbd850b44c88/moonshine/model.py#L348-L466).

    def __init__(
        self,
        hidden_dim,
        intermediate_dim,
        num_heads,
        feedforward_expansion_factor=4,
        use_swiglu_activation=True,
        pad_head_dim_to_multiple_of=None,
        initializer_range=0.02,
        attention_bias=False,
        attention_dropout=0.0,
        dtype=None,
        **kwargs,
    ):
        kwargs.pop("dropout", None)
        kwargs.pop("activation", None)
        kwargs.pop("kernel_initializer", None)
        self.kernel_initializer = moonshine_kernel_initializer(
            initializer_range=initializer_range
        )
        super().__init__(
            intermediate_dim=intermediate_dim,
            num_heads=num_heads,
            dropout=attention_dropout,
            activation="gelu" if use_swiglu_activation else "silu",
            kernel_initializer=clone_initializer(self.kernel_initializer),
            dtype=dtype,
            **kwargs,
        )
        self.initializer_range = initializer_range
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.num_heads = num_heads
        self.feedforward_expansion_factor = feedforward_expansion_factor
        self.use_swiglu_activation = use_swiglu_activation
        self.pad_head_dim_to_multiple_of = pad_head_dim_to_multiple_of
        self.attention_dropout = attention_dropout
        self.attention_bias = attention_bias

        self.head_dim = hidden_dim // num_heads
        if pad_head_dim_to_multiple_of is not None:
            self.head_dim = (
                (self.head_dim + pad_head_dim_to_multiple_of - 1)
                // pad_head_dim_to_multiple_of
            ) * pad_head_dim_to_multiple_of

        self.norm1 = keras.layers.LayerNormalization(
            axis=-1,
            epsilon=1e-5,
            center=False,
            scale=True,
            dtype=self.dtype,
        )
        self.self_attention = MoonshineMultiHeadAttention(
            num_heads=num_heads,
            key_dim=self.head_dim,
            use_bias=False,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            attention_bias=attention_bias,
            attention_dropout=attention_dropout,
            use_causal_mask=True,
            apply_rotary_embedding=True,
            dtype=self.dtype,
        )
        self.norm2 = keras.layers.LayerNormalization(
            axis=-1,
            epsilon=1e-5,
            center=False,
            scale=True,
            dtype=self.dtype,
        )
        self.cross_attention = MoonshineMultiHeadAttention(
            num_heads=num_heads,
            key_dim=self.head_dim,
            use_bias=False,
            kernel_initializer=clone_initializer(self.kernel_initializer),
            attention_bias=attention_bias,
            attention_dropout=attention_dropout,
            use_causal_mask=False,
            apply_rotary_embedding=False,
            dtype=self.dtype,
        )
        self.norm3 = keras.layers.LayerNormalization(
            axis=-1,
            epsilon=1e-5,
            center=False,
            scale=True,
            dtype=self.dtype,
        )
        self.ff = MoonshineMLP(
            hidden_dim=hidden_dim,
            feedforward_expansion_factor=feedforward_expansion_factor,
            use_swiglu_activation=use_swiglu_activation,
            initializer_range=initializer_range,
            dtype=self.dtype,
        )

    def build(self, decoder_sequence_shape, encoder_sequence_shape=None):
        if encoder_sequence_shape is None:
            raise ValueError(
                "Encoder sequence shape must be provided for "
                "MoonshineDecoderBlock."
            )
        context_shape = encoder_sequence_shape  # Shape of context

        # Build sublayers.
        self.norm1.build(decoder_sequence_shape)
        self.norm2.build(decoder_sequence_shape)
        self.norm3.build(decoder_sequence_shape)

        self.self_attention.build(
            query_shape=decoder_sequence_shape,
            key_shape=decoder_sequence_shape,
            value_shape=decoder_sequence_shape,
        )

        self.cross_attention.build(
            query_shape=decoder_sequence_shape,
            key_shape=context_shape,
            value_shape=context_shape,
        )

        self.ff.build(decoder_sequence_shape)
        self.built = True

    def _compute_self_attention_mask(
        self,
        decoder_sequence,
        decoder_padding_mask,
        self_attention_cache=None,
        self_attention_cache_update_index=None,
    ):
        decoder_mask = merge_padding_and_attention_mask(
            inputs=decoder_sequence,
            padding_mask=decoder_padding_mask,
            attention_mask=None,
        )
        if self.self_attention.use_causal_mask:
            batch_size = keras.ops.shape(decoder_sequence)[0]
            output_length = keras.ops.shape(decoder_sequence)[1]
            current_cache_update_index = (
                0
                if self_attention_cache_update_index is None
                else self_attention_cache_update_index
            )
            if self_attention_cache is not None:
                input_length = keras.ops.shape(self_attention_cache)[2]
            else:
                input_length = output_length
            causal_mask = compute_causal_mask(
                batch_size,
                input_length,
                output_length,
                current_cache_update_index,
            )
            return (
                keras.ops.minimum(decoder_mask, causal_mask)
                if decoder_mask is not None
                else causal_mask
            )
        return decoder_mask

    def call(
        self,
        decoder_sequence,
        encoder_sequence,
        rotary_embedding,
        encoder_attention_mask=None,
        decoder_padding_mask=None,
        encoder_padding_mask=None,
        self_attention_cache=None,
        self_attention_cache_update_index=None,
        cross_attention_cache=None,
        cross_attention_cache_update_index=None,
        training=None,
    ):
        x = decoder_sequence
        context = encoder_sequence
        has_self_attention_cache = self_attention_cache is not None
        has_cross_attention_cache = cross_attention_cache is not None

        self_attention_mask = self._compute_self_attention_mask(
            decoder_sequence=x,
            decoder_padding_mask=decoder_padding_mask,
            self_attention_cache=self_attention_cache,
            self_attention_cache_update_index=self_attention_cache_update_index,
        )

        # Self attention block.
        residual = x
        x_norm1 = self.norm1(x)
        x_self_attn = self.self_attention(
            query=x_norm1,
            key=x_norm1,
            value=x_norm1,
            rotary_embedding=rotary_embedding,
            cache=self_attention_cache,
            cache_update_index=self_attention_cache_update_index,
            attention_mask=self_attention_mask,
            training=training,
        )
        if has_self_attention_cache:
            x_self_attn, self_attention_cache = x_self_attn
        x = x_self_attn + residual
        # Cross attention block.
        residual = x
        x_norm2 = self.norm2(x)
        cross_attention_mask = merge_padding_and_attention_mask(
            inputs=encoder_sequence,
            padding_mask=encoder_padding_mask,
            attention_mask=encoder_attention_mask,
        )
        x_cross_attn = self.cross_attention(
            query=x_norm2,
            key=context,
            value=context,
            cache=cross_attention_cache,
            cache_update_index=cross_attention_cache_update_index,
            attention_mask=cross_attention_mask,
            training=training,
        )
        if has_cross_attention_cache:
            x_cross_attn, cross_attention_cache = x_cross_attn
        x = x_cross_attn + residual
        residual = x
        x_norm3 = self.norm3(x)
        x_ff = self.ff(x_norm3)
        x = x_ff + residual

        if has_self_attention_cache:
            return x, self_attention_cache
        return x

    def compute_output_shape(
        self, decoder_sequence_shape, encoder_sequence_shape=None
    ):
        return decoder_sequence_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "num_heads": self.num_heads,
                "feedforward_expansion_factor": self.feedforward_expansion_factor,  # noqa: E501
                "use_swiglu_activation": self.use_swiglu_activation,
                "pad_head_dim_to_multiple_of": self.pad_head_dim_to_multiple_of,  # noqa: E501
                "initializer_range": self.initializer_range,
                "attention_bias": self.attention_bias,
                "attention_dropout": self.attention_dropout,
                "dtype": self.dtype,
            }
        )
        return config



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)
from keras_hub.src.models.vit.vit_backbone import ViTBackbone
from keras_hub.src.models.vit.vit_image_converter import ViTImageConverter


@keras_hub_export("keras_hub.models.ViTImageClassifierPreprocessor")
class ViTImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = ViTBackbone
    image_converter_cls = ViTImageConverter



import numpy as np

from keras_hub.src.models.resnet.resnet_backbone import ResNetBackbone

backbone_cls = ResNetBackbone


def convert_backbone_config(timm_config):
    timm_architecture = timm_config["architecture"]

    if "resnetv2_" in timm_architecture:
        use_pre_activation = True
    else:
        use_pre_activation = False

    if timm_architecture == "resnet18":
        stackwise_num_blocks = [2, 2, 2, 2]
        block_type = "basic_block"
    elif timm_architecture == "resnet26":
        stackwise_num_blocks = [2, 2, 2, 2]
        block_type = "bottleneck_block"
    elif timm_architecture == "resnet34":
        stackwise_num_blocks = [3, 4, 6, 3]
        block_type = "basic_block"
    elif timm_architecture in ("resnet50", "resnetv2_50"):
        stackwise_num_blocks = [3, 4, 6, 3]
        block_type = "bottleneck_block"
    elif timm_architecture in ("resnet101", "resnetv2_101"):
        stackwise_num_blocks = [3, 4, 23, 3]
        block_type = "bottleneck_block"
    elif timm_architecture in ("resnet152", "resnetv2_152"):
        stackwise_num_blocks = [3, 8, 36, 3]
        block_type = "bottleneck_block"
    else:
        raise ValueError(
            f"Currently, the architecture {timm_architecture} is not supported."
        )

    return dict(
        stackwise_num_filters=[64, 128, 256, 512],
        stackwise_num_blocks=stackwise_num_blocks,
        stackwise_num_strides=[1, 2, 2, 2],
        block_type=block_type,
        use_pre_activation=use_pre_activation,
        input_conv_filters=[64],
        input_conv_kernel_sizes=[7],
    )


def convert_weights(backbone, loader, timm_config):
    def port_conv2d(keras_layer_name, hf_weight_prefix):
        loader.port_weight(
            backbone.get_layer(keras_layer_name).kernel,
            hf_weight_key=f"{hf_weight_prefix}.weight",
            hook_fn=lambda x, _: np.transpose(x, (2, 3, 1, 0)),
        )

    def port_batch_normalization(keras_layer_name, hf_weight_prefix):
        loader.port_weight(
            backbone.get_layer(keras_layer_name).gamma,
            hf_weight_key=f"{hf_weight_prefix}.weight",
        )
        loader.port_weight(
            backbone.get_layer(keras_layer_name).beta,
            hf_weight_key=f"{hf_weight_prefix}.bias",
        )
        loader.port_weight(
            backbone.get_layer(keras_layer_name).moving_mean,
            hf_weight_key=f"{hf_weight_prefix}.running_mean",
        )
        loader.port_weight(
            backbone.get_layer(keras_layer_name).moving_variance,
            hf_weight_key=f"{hf_weight_prefix}.running_var",
        )

    version = "v1" if not backbone.use_pre_activation else "v2"
    block_type = backbone.block_type

    # Stem
    if version == "v1":
        port_conv2d("conv1_conv", "conv1")
        port_batch_normalization("conv1_bn", "bn1")
    else:
        port_conv2d("conv1_conv", "stem.conv")

    # Stages
    num_stacks = len(backbone.stackwise_num_filters)
    for stack_index in range(num_stacks):
        for block_idx in range(backbone.stackwise_num_blocks[stack_index]):
            if version == "v1":
                keras_name = f"stack{stack_index}_block{block_idx}"
                hf_name = f"layer{stack_index + 1}.{block_idx}"
            else:
                keras_name = f"stack{stack_index}_block{block_idx}"
                hf_name = f"stages.{stack_index}.blocks.{block_idx}"

            if version == "v1":
                if block_idx == 0 and (
                    block_type == "bottleneck_block" or stack_index > 0
                ):
                    port_conv2d(
                        f"{keras_name}_0_conv", f"{hf_name}.downsample.0"
                    )
                    port_batch_normalization(
                        f"{keras_name}_0_bn", f"{hf_name}.downsample.1"
                    )
                port_conv2d(f"{keras_name}_1_conv", f"{hf_name}.conv1")
                port_batch_normalization(f"{keras_name}_1_bn", f"{hf_name}.bn1")
                port_conv2d(f"{keras_name}_2_conv", f"{hf_name}.conv2")
                port_batch_normalization(f"{keras_name}_2_bn", f"{hf_name}.bn2")
                if block_type == "bottleneck_block":
                    port_conv2d(f"{keras_name}_3_conv", f"{hf_name}.conv3")
                    port_batch_normalization(
                        f"{keras_name}_3_bn", f"{hf_name}.bn3"
                    )
            else:
                if block_idx == 0 and (
                    block_type == "bottleneck_block" or stack_index > 0
                ):
                    port_conv2d(
                        f"{keras_name}_0_conv", f"{hf_name}.downsample.conv"
                    )
                port_batch_normalization(
                    f"{keras_name}_pre_activation_bn", f"{hf_name}.norm1"
                )
                port_conv2d(f"{keras_name}_1_conv", f"{hf_name}.conv1")
                port_batch_normalization(
                    f"{keras_name}_1_bn", f"{hf_name}.norm2"
                )
                port_conv2d(f"{keras_name}_2_conv", f"{hf_name}.conv2")
                if block_type == "bottleneck_block":
                    port_batch_normalization(
                        f"{keras_name}_2_bn", f"{hf_name}.norm3"
                    )
                    port_conv2d(f"{keras_name}_3_conv", f"{hf_name}.conv3")

    # Post
    if version == "v2":
        port_batch_normalization("post_bn", "norm")


def convert_head(task, loader, timm_config):
    v2 = "resnetv2_" in timm_config["architecture"]
    prefix = "head.fc." if v2 else "fc."
    loader.port_weight(
        task.output_dense.kernel,
        hf_weight_key=prefix + "weight",
        hook_fn=lambda x, _: np.transpose(np.squeeze(x)),
    )
    loader.port_weight(
        task.output_dense.bias,
        hf_weight_key=prefix + "bias",
    )



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.task import Task


class Multiplier(keras.layers.Layer):
    def __init__(self, multiplier=None, **kwargs):
        super().__init__(**kwargs)
        self.multiplier = float(multiplier) if multiplier is not None else None

    def call(self, inputs):
        if self.multiplier is not None:
            inputs = keras.ops.multiply(inputs, self.multiplier)
        return inputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "multiplier": self.multiplier,
            }
        )
        return config


@keras_hub_export("keras_hub.models.DepthEstimator")
class DepthEstimator(Task):
    """Base class for all depth estimation tasks.

    `DepthEstimator` tasks wrap a `keras_hub.models.Backbone` and
    a `keras_hub.models.Preprocessor` to create a model that can be used for
    depth estimation.

    To fine-tune with `fit()`, pass a dataset containing tuples of `(x, y)`
    labels where `x` is a RGB image and `y` is a depth map. All `DepthEstimator`
    tasks include a `from_preset()` constructor which can be used to load a
    pre-trained config and weights.

    Args:
        backbone: A `keras_hub.models.Backbone` instance or a `keras.Model`.
        preprocessor: `None`, a `keras_hub.models.Preprocessor` instance,
            a `keras.Layer` instance, or a callable. If `None` no preprocessing
            will be applied to the inputs.
        depth_estimation_type: `"relative"` or `"metric"`. The type of depth map
            to use. `"relative"` depth maps are up-to-scale, while `"metric"`
            depth maps have metric meaning (e.g. in meters). Defaults to
            `"relative"`.
        min_depth: An float representing the minimum depth value. This value can
            be used to filter out invalid depth values during training. Defaults
            to `keras.config.epsilon()`.
        max_depth: An optional float representing the maximum depth value. This
            value can be used to filter out invalid depth values during
            training. When `depth_estimation_type="metric"`, the model's output
            will be scaled to the range `[0, max_depth]`.

    Examples:

    Call `predict()` to run inference.
    ```python
    # Load preset and train
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    depth_estimator = keras_hub.models.DepthEstimator.from_preset(
        "depth_anything_v2_small"
    )
    depth_estimator.predict(images)
    ```

    Call `fit()` on a single batch.
    ```python
    # Load preset and train
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    depths = np.random.uniform(0, 10, size=(2, 224, 224))
    depth_estimator = keras_hub.models.DepthEstimator.from_preset(
        "depth_anything_v2_small",
        depth_estimation_type="metric",
        max_depth=10.0,
    )
    depth_estimator.fit(x=images, y=depths, batch_size=2)
    ```

    Call `fit()` with custom loss, optimizer and backbone.
    ```python
    depth_estimator = keras_hub.models.DepthEstimator.from_preset(
        "depth_anything_v2_small",
        depth_estimation_type="metric",
        max_depth=10.0,
    )
    depth_estimator.compile(
        loss=keras.losses.MeanSquaredError(),
        optimizer=keras.optimizers.Adam(5e-5),
    )
    depth_estimator.backbone.trainable = False
    depth_estimator.fit(x=images, y=depths, batch_size=2)
    ```

    Custom backbone.
    ```python
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    depths = np.random.uniform(0, 10, size=(2, 224, 224))
    image_encoder = keras_hub.models.DINOV2Backbone.from_preset("dinov2_small")
    backbone = keras_hub.models.DepthAnythingBackbone(
        image_encoder=image_encoder,
        patch_size=image_encoder.patch_size,
        backbone_hidden_dim=image_encoder.hidden_dim,
        reassemble_factors=[4, 2, 1, 0.5],
        neck_hidden_dims=[48, 96, 192, 384],
        fusion_hidden_dim=64,
        head_hidden_dim=32,
        head_in_index=-1,
    )
    depth_estimator = keras_hub.models.DepthEstimator(
        backbone=backbone,
        depth_estimation_type="metric",
        max_depth=10.0,
    )
    depth_estimator.fit(x=images, y=depths, batch_size=2)
    ```
    """

    def __init__(
        self,
        backbone,
        depth_estimation_type,
        min_depth=keras.config.epsilon(),
        max_depth=None,
        preprocessor=None,
        **kwargs,
    ):
        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        if depth_estimation_type == "relative":
            self.output_activation = keras.layers.ReLU(
                dtype=backbone.dtype_policy,
                name="output_activation",
            )
        elif depth_estimation_type == "metric":
            self.output_activation = keras.layers.Activation(
                activation="sigmoid",
                dtype=backbone.dtype_policy,
                name="output_activation",
            )
        else:
            raise ValueError(
                "`depth_estimation_type` should be either `'relative'` or "
                "`'metric'`. "
                f"Received: depth_estimation_type={depth_estimation_type}."
            )
        if max_depth is not None and depth_estimation_type != "metric":
            raise ValueError(
                "`max_depth` should only be set when "
                "`depth_estimation_type='metric'`. "
                f"Received: depth_estimation_type={depth_estimation_type}, "
                f"max_depth={max_depth}."
            )
        self.multiplier = Multiplier(
            multiplier=max_depth, dtype=backbone.dtype_policy, name="multiplier"
        )
        self.depths = keras.layers.Identity(
            dtype=backbone.dtype_policy, name="depths"
        )

        # === Config ===
        self.depth_estimation_type = depth_estimation_type
        self.min_depth = float(min_depth) if min_depth is not None else None
        self.max_depth = float(max_depth) if max_depth is not None else None

        # === Functional Model ===
        inputs = self.backbone.input
        depths = self.backbone(inputs)
        depths = self.output_activation(depths)
        depths = self.multiplier(depths)
        depths = self.depths(depths)
        outputs = {"depths": depths}
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

    def get_config(self):
        # Backbone serialized in `super`
        config = super().get_config()
        config.update(
            {
                "depth_estimation_type": self.depth_estimation_type,
                "min_depth": self.min_depth,
                "max_depth": self.max_depth,
            }
        )
        return config

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        *,
        metrics="auto",
        **kwargs,
    ):
        """Configures the `DepthEstimator` task for training.

        The `DepthEstimator` task extends the default compilation signature of
        `keras.Model.compile` with defaults for `optimizer`, `loss`, and
        `metrics`. To override these defaults, pass any value
        to these arguments during compilation.

        Args:
            optimizer: `"auto"`, an optimizer name, or a `keras.Optimizer`
                instance. Defaults to `"auto"`, which uses the default optimizer
                for the given model and task. See `keras.Model.compile` and
                `keras.optimizers` for more info on possible `optimizer` values.
            loss: `"auto"`, a loss name, or a `keras.losses.Loss` instance.
                Defaults to `"auto"`, where a `keras.losses.MeanSquaredError`
                loss will be applied for the depth estimation task. See
                `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            metrics: `"auto"`, or a dict of metrics to be evaluated by
                the model during training and testing. Defaults to `"auto"`,
                where a `keras.metrics.RootMeanSquaredError` will be applied to
                track the accuracy of the model during training. See
                `keras.Model.compile` and `keras.metrics` for more info on
                possible `metrics` values.
            **kwargs: See `keras.Model.compile` for a full list of arguments
                supported by the compile method.
        """
        if optimizer == "auto":
            optimizer = keras.optimizers.AdamW(5e-5)
        if loss == "auto":
            loss = {"depths": keras.losses.MeanSquaredError()}
        if metrics == "auto":
            metrics = {"depths": keras.metrics.RootMeanSquaredError()}
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            **kwargs,
        )



import keras
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.masked_lm import MaskedLM
from keras_hub.src.models.roformer_v2.roformer_v2_backbone import (
    RoformerV2Backbone,
)
from keras_hub.src.models.roformer_v2.roformer_v2_masked_lm_preprocessor import (  # noqa: E501
    RoformerV2MaskedLMPreprocessor,
)


class RoformerV2MaskedLMHead(keras.layers.Layer):
    def __init__(
        self,
        vocabulary_size=None,
        token_embedding=None,
        activation=None,
        **kwargs,
    ):
        super().__init__(**kwargs, autocast=False)

        self.token_embedding = token_embedding
        self.activation = keras.activations.get(activation)

        if vocabulary_size and vocabulary_size != token_embedding.input_dim:
            raise ValueError(
                "`vocabulary_size` should match the input dimension of the "
                "of `token_embedding`. Received: "
                f"`vocabulary_size={vocabulary_size}`, "
                f"`token_embedding.input_dim={token_embedding.input_dim}`"
            )
        self.vocabulary_size = token_embedding.input_dim

    def call(self, inputs, mask_positions):
        if keras.config.backend() == "tensorflow":
            import tensorflow as tf

            # On the tf backend, we need to work around an issue with dynamic
            # shape broadcasting in take_along_axis.
            x = tf.gather(inputs, mask_positions, batch_dims=1)
        else:
            # Gather the encoded tokens at the masked indices.
            mask_positions = ops.expand_dims(mask_positions, axis=-1)
            x = ops.take_along_axis(inputs, mask_positions, axis=1)

        outputs = self.token_embedding(x, reverse=True)

        # Apply a final activation.
        if self.activation is not None:
            outputs = self.activation(outputs)

        outputs = ops.cast(outputs, "float32")
        return outputs

    def get_config(self):
        config = super().get_config()
        embedding_config = None
        if self.token_embedding:
            embedding_config = keras.layers.serialize(self.token_embedding)
        config.update(
            {
                "token_embedding": embedding_config,
                "vocabulary_size": self.vocabulary_size,
                "activation": keras.activations.serialize(self.activation),
            }
        )
        return config


@keras_hub_export("keras_hub.models.RoformerV2MaskedLM")
class RoformerV2MaskedLM(MaskedLM):
    """An end-to-end Roformer model for the masked language modeling task.

    This model will train RoformerV2 on a masked language modeling task.
    The model will predict labels for a number of masked tokens in the
    input data. For usage of this model with pre-trained weights, see the
    `from_preset()` constructor.

    This model can optionally be configured with a `preprocessor` layer, in
    which case inputs can be raw string features during `fit()`, `predict()`,
    and `evaluate()`. Inputs will be tokenized and dynamically masked during
    training and evaluation. This is done by default when creating the model
    with `from_preset()`.

    Disclaimer: Pre-trained models are provided on an "as is" basis, without
    warranties or conditions of any kind.

    Args:
        backbone: A `keras_hub.models.RoformerV2Backbone` instance.
        preprocessor: A `keras_hub.models.RoformerV2MaskedLMPreprocessor` or
            `None`. If `None`, this model will not apply preprocessing, and
            inputs should be preprocessed before calling the model.

    Examples:

    Raw string data.
    ```python
    features = ["The quick brown fox jumped.", "I forgot my homework."]

    # Pretrained language model.
    masked_lm = keras_hub.models.RoformerV2MaskedLM.from_preset(
        "roformer_v2_base_zh",
    )
    masked_lm.fit(x=features, batch_size=2)

    # Re-compile (e.g., with a new learning rate).
    masked_lm.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=keras.optimizers.Adam(5e-5),
        jit_compile=True,
    )
    # Access backbone programmatically (e.g., to change `trainable`).
    masked_lm.backbone.trainable = False
    # Fit again.
    masked_lm.fit(x=features, batch_size=2)
    ```

    Preprocessed integer data.
    ```python
    # Create preprocessed batch where 0 is the mask token.
    features = {
        "token_ids": np.array([[1, 2, 0, 4, 0, 6, 7, 8]] * 2),
        "padding_mask": np.array([[1, 1, 1, 1, 1, 1, 1, 1]] * 2),
        "mask_positions": np.array([[2, 4]] * 2),
        "segment_ids": np.array([[0, 0, 0, 0, 0, 0, 0, 0]] * 2)
    }
    # Labels are the original masked values.
    labels = [[3, 5]] * 2

    masked_lm = keras_hub.models.RoformerV2MaskedLM.from_preset(
        "roformer_v2_base_zh",
        preprocessor=None,
    )
    masked_lm.fit(x=features, y=labels, batch_size=2)
    ```
    """

    backbone_cls = RoformerV2Backbone
    preprocessor_cls = RoformerV2MaskedLMPreprocessor

    def __init__(
        self,
        backbone,
        preprocessor=None,
        **kwargs,
    ):
        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        self.masked_lm_head = RoformerV2MaskedLMHead(
            vocabulary_size=backbone.vocabulary_size,
            token_embedding=backbone.token_embedding,
            name="mlm_head",
        )

        # === Functional Model ===
        inputs = {
            **backbone.input,
            "mask_positions": keras.Input(
                shape=(None,), dtype="int32", name="mask_positions"
            ),
        }
        backbone_outputs = backbone(backbone.input)
        outputs = self.masked_lm_head(
            backbone_outputs, inputs["mask_positions"]
        )
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.albert.albert_backbone import AlbertBackbone
from keras_hub.src.tokenizers.sentence_piece_tokenizer import (
    SentencePieceTokenizer,
)


@keras_hub_export(
    [
        "keras_hub.tokenizers.AlbertTokenizer",
        "keras_hub.models.AlbertTokenizer",
    ]
)
class AlbertTokenizer(SentencePieceTokenizer):
    """ALBERT tokenizer layer based on SentencePiece.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_hub.tokenizers.SentencePieceTokenizer`. Unlike the
    underlying tokenizer, it will check for all special tokens needed by
    ALBERT models and provides a `from_preset()` method to automatically
    download a matching vocabulary for a ALBERT preset.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.

    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Args:
        proto: Either a `string` path to a SentencePiece proto file, or a
            `bytes` object with a serialized SentencePiece proto. See the
            [SentencePiece repository](https://github.com/google/sentencepiece)
            for more details on the format.

    Examples:

    ```python
    # Unbatched input.
    tokenizer = keras_hub.models.AlbertTokenizer.from_preset(
        "albert_base_en_uncased",
    )
    tokenizer("The quick brown fox jumped.")

    # Batched input.
    tokenizer(["The quick brown fox jumped.", "The fox slept."])

    # Detokenization.
    tokenizer.detokenize(tokenizer("The quick brown fox jumped."))

    # Custom vocabulary.
    bytes_io = io.BytesIO()
    ds = tf.data.Dataset.from_tensor_slices(["The quick brown fox jumped."])
    sentencepiece.SentencePieceTrainer.train(
        sentence_iterator=ds.as_numpy_iterator(),
        model_writer=bytes_io,
        vocab_size=10,
        model_type="WORD",
        pad_id=0,
        unk_id=1,
        bos_id=2,
        eos_id=3,
        pad_piece="<pad>",
        unk_piece="<unk>",
        bos_piece="[CLS]",
        eos_piece="[SEP]",
        user_defined_symbols="[MASK]",
    )
    tokenizer = keras_hub.models.AlbertTokenizer(
        proto=bytes_io.getvalue(),
    )
    tokenizer("The quick brown fox jumped.")
    ```
    """

    backbone_cls = AlbertBackbone

    def __init__(self, proto, **kwargs):
        self._add_special_token("[CLS]", "cls_token")
        self._add_special_token("[SEP]", "sep_token")
        self._add_special_token("<pad>", "pad_token")
        self._add_special_token("[MASK]", "mask_token")
        # Also add `tokenizer.start_token` and `tokenizer.end_token` for
        # compatibility with other tokenizers.
        self._add_special_token("[CLS]", "start_token")
        self._add_special_token("[SEP]", "end_token")
        super().__init__(proto=proto, **kwargs)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.deeplab_v3.deeplab_v3_backbone import (
    DeepLabV3Backbone,
)
from keras_hub.src.models.deeplab_v3.deeplab_v3_image_converter import (
    DeepLabV3ImageConverter,
)
from keras_hub.src.models.image_segmenter_preprocessor import (
    ImageSegmenterPreprocessor,
)


@keras_hub_export("keras_hub.models.DeepLabV3ImageSegmenterPreprocessor")
class DeepLabV3ImageSegmenterPreprocessor(ImageSegmenterPreprocessor):
    backbone_cls = DeepLabV3Backbone
    image_converter_cls = DeepLabV3ImageConverter



"""PARSeq preset configurations."""

backbone_presets = {
    "parseq": {
        "metadata": {
            "description": (
                "Permuted autoregressive sequence (PARSeq) base "
                "model for scene text recognition"
            ),
            "params": 23_832_671,
            "path": "parseq",
        },
        "kaggle_handle": "kaggle://keras/parseq/keras/parseq/1",
    }
}



"""ESM model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "esm2_t6_8M": {
        "metadata": {
            "description": (
                "6 transformer layers version of the ESM-2 protein language "
                "model, trained on the UniRef50 clustered protein sequence "
                "dataset."
            ),
            "params": 7_408_960,
            "path": "esm",
        },
        "kaggle_handle": "kaggle://keras/esm-2/keras/esm2_t6_8M/1",
    },
    "esm2_t12_35M": {
        "metadata": {
            "description": (
                "12 transformer layers version of the ESM-2 protein language "
                "model, trained on the UniRef50 clustered protein sequence "
                "dataset."
            ),
            "params": 33_269_280,
            "path": "esm",
        },
        "kaggle_handle": "kaggle://keras/esm-2/keras/esm2_t12_35M/1",
    },
    "esm2_t30_150M": {
        "metadata": {
            "description": (
                "30 transformer layers version of the ESM-2 protein language "
                "model, trained on the UniRef50 clustered protein sequence "
                "dataset."
            ),
            "params": 147_728_000,
            "path": "esm",
        },
        "kaggle_handle": "kaggle://keras/esm-2/keras/esm2_t30_150M/1",
    },
    "esm2_t33_650M": {
        "metadata": {
            "description": (
                "33 transformer layers version of the ESM-2 protein language "
                "model, trained on the UniRef50 clustered protein sequence "
                "dataset."
            ),
            "params": 649_400_320,
            "path": "esm",
        },
        "kaggle_handle": "kaggle://keras/esm-2/keras/esm2_t33_650M/1",
    },
}



from keras_hub.src.models.deit.deit_backbone import DeiTBackbone
from keras_hub.src.models.deit.deit_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, DeiTBackbone)



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.task import Task


@keras_hub_export("keras_hub.models.ImageClassifier")
class ImageClassifier(Task):
    """Base class for all image classification tasks.

    `ImageClassifier` tasks wrap a `keras_hub.models.Backbone` and
    a `keras_hub.models.Preprocessor` to create a model that can be used for
    image classification. `ImageClassifier` tasks take an additional
    `num_classes` argument, controlling the number of predicted output classes.

    To fine-tune with `fit()`, pass a dataset containing tuples of `(x, y)`
    labels where `x` is a string and `y` is a integer from `[0, num_classes)`.
    All `ImageClassifier` tasks include a `from_preset()` constructor which can
    be used to load a pre-trained config and weights.

    Args:
        backbone: A `keras_hub.models.Backbone` instance or a `keras.Model`.
        num_classes: int. The number of classes to predict.
        preprocessor: `None`, a `keras_hub.models.Preprocessor` instance,
            a `keras.Layer` instance, or a callable. If `None` no preprocessing
            will be applied to the inputs.
        pooling: `"avg"` or `"max"`. The type of pooling to apply on backbone
            output. Defaults to average pooling.
        activation: `None`, str, or callable. The activation function to use on
            the `Dense` layer. Set `activation=None` to return the output
            logits. Defaults to `"softmax"`.
        head_dtype: `None`, str, or `keras.mixed_precision.DTypePolicy`. The
            dtype to use for the classification head's computations and weights.

    Examples:

    Call `predict()` to run inference.
    ```python
    # Load preset and train
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    classifier = keras_hub.models.ImageClassifier.from_preset(
        "resnet_50_imagenet"
    )
    classifier.predict(images)
    ```

    Call `fit()` on a single batch.
    ```python
    # Load preset and train
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    labels = [0, 3]
    classifier = keras_hub.models.ImageClassifier.from_preset(
        "resnet_50_imagenet"
    )
    classifier.fit(x=images, y=labels, batch_size=2)
    ```

    Call `fit()` with custom loss, optimizer and backbone.
    ```python
    classifier = keras_hub.models.ImageClassifier.from_preset(
        "resnet_50_imagenet"
    )
    classifier.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=keras.optimizers.Adam(5e-5),
    )
    classifier.backbone.trainable = False
    classifier.fit(x=images, y=labels, batch_size=2)
    ```

    Custom backbone.
    ```python
    images = np.random.randint(0, 256, size=(2, 224, 224, 3))
    labels = [0, 3]
    backbone = keras_hub.models.ResNetBackbone(
        stackwise_num_filters=[64, 64, 64],
        stackwise_num_blocks=[2, 2, 2],
        stackwise_num_strides=[1, 2, 2],
        block_type="basic_block",
        use_pre_activation=True,
        pooling="avg",
    )
    classifier = keras_hub.models.ImageClassifier(
        backbone=backbone,
        num_classes=4,
    )
    classifier.fit(x=images, y=labels, batch_size=2)
    ```
    """

    def __init__(
        self,
        backbone,
        num_classes,
        preprocessor=None,
        pooling="avg",
        activation=None,
        dropout=0.0,
        head_dtype=None,
        **kwargs,
    ):
        head_dtype = head_dtype or backbone.dtype_policy
        data_format = getattr(backbone, "data_format", None)

        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        if pooling == "avg":
            self.pooler = keras.layers.GlobalAveragePooling2D(
                data_format,
                dtype=head_dtype,
                name="pooler",
            )
        elif pooling == "max":
            self.pooler = keras.layers.GlobalMaxPooling2D(
                data_format,
                dtype=head_dtype,
                name="pooler",
            )
        else:
            raise ValueError(
                "Unknown `pooling` type. Polling should be either `'avg'` or "
                f"`'max'`. Received: pooling={pooling}."
            )
        self.output_dropout = keras.layers.Dropout(
            dropout,
            dtype=head_dtype,
            name="output_dropout",
        )
        self.output_dense = keras.layers.Dense(
            num_classes,
            activation=activation,
            dtype=head_dtype,
            name="predictions",
        )

        # === Functional Model ===
        inputs = self.backbone.input
        x = self.backbone(inputs)
        x = self.pooler(x)
        x = self.output_dropout(x)
        outputs = self.output_dense(x)
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

        # === Config ===
        self.num_classes = num_classes
        self.activation = activation
        self.pooling = pooling
        self.dropout = dropout

    def get_config(self):
        # Backbone serialized in `super`
        config = super().get_config()
        config.update(
            {
                "num_classes": self.num_classes,
                "pooling": self.pooling,
                "activation": self.activation,
                "dropout": self.dropout,
            }
        )
        return config

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        *,
        metrics="auto",
        **kwargs,
    ):
        """Configures the `ImageClassifier` task for training.

        The `ImageClassifier` task extends the default compilation signature of
        `keras.Model.compile` with defaults for `optimizer`, `loss`, and
        `metrics`. To override these defaults, pass any value
        to these arguments during compilation.

        Args:
            optimizer: `"auto"`, an optimizer name, or a `keras.Optimizer`
                instance. Defaults to `"auto"`, which uses the default optimizer
                for the given model and task. See `keras.Model.compile` and
                `keras.optimizers` for more info on possible `optimizer` values.
            loss: `"auto"`, a loss name, or a `keras.losses.Loss` instance.
                Defaults to `"auto"`, where a
                `keras.losses.SparseCategoricalCrossentropy` loss will be
                applied for the classification task. See
                `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            metrics: `"auto"`, or a list of metrics to be evaluated by
                the model during training and testing. Defaults to `"auto"`,
                where a `keras.metrics.SparseCategoricalAccuracy` will be
                applied to track the accuracy of the model during training.
                See `keras.Model.compile` and `keras.metrics` for
                more info on possible `metrics` values.
            **kwargs: See `keras.Model.compile` for a full list of arguments
                supported by the compile method.
        """
        if optimizer == "auto":
            optimizer = keras.optimizers.Adam(5e-5)
        if loss == "auto":
            activation = getattr(self, "activation", None)
            activation = keras.activations.get(activation)
            from_logits = activation != keras.activations.softmax
            loss = keras.losses.SparseCategoricalCrossentropy(from_logits)
        if metrics == "auto":
            metrics = [keras.metrics.SparseCategoricalAccuracy()]
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            **kwargs,
        )



from keras import ops


def rotate_half(x):
    x1 = x[..., : ops.shape(x)[-1] // 2]
    x2 = x[..., ops.shape(x)[-1] // 2 :]
    return ops.concatenate((-x2, x1), axis=-1)


def apply_rotary_pos_emb(q, k, cos, sin, expansion_axis=1):
    cos = ops.expand_dims(cos, expansion_axis)
    sin = ops.expand_dims(sin, expansion_axis)
    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed


def apply_rotary_pos_single(tensor, cos, sin, expansion_axis=1):
    cos = ops.expand_dims(cos, expansion_axis)
    sin = ops.expand_dims(sin, expansion_axis)
    tensor_embed = (tensor * cos) + (rotate_half(tensor) * sin)
    return tensor_embed


def repeat_kv(hidden_states, n_rep):
    batch, num_key_value_heads, slen, head_dim = ops.shape(hidden_states)
    if n_rep == 1:
        return hidden_states
    hidden_states = ops.expand_dims(hidden_states, axis=2)
    target_shape = (batch, num_key_value_heads, n_rep, slen, head_dim)
    hidden_states = ops.broadcast_to(hidden_states, target_shape)
    return ops.reshape(
        hidden_states, [batch, num_key_value_heads * n_rep, slen, head_dim]
    )


def rope_init(rope_theta, partial_rotary_factor, head_dim):
    """Initialize RoPE (Rotary Position Embedding) parameters.

    Args:
        rope_theta: float. The theta value for RoPE.
        partial_rotary_factor: float. The factor for partial rotary embedding.
        head_dim: int. The dimension of each attention head.

    Returns:
        A tuple of (inv_freq, attention_scaling) where inv_freq is the inverse
        frequency tensor and attention_scaling is the scaling factor.
    """
    base = rope_theta
    dim = int(head_dim * partial_rotary_factor)

    inv_freq = 1.0 / (
        ops.power(base, ops.arange(0, dim, 2, dtype="float32") / dim)
    )
    attention_scaling = 1.0
    return inv_freq, attention_scaling



import numpy as np

from keras_hub.src.models.smollm3.smollm3_backbone import SmolLM3Backbone
from keras_hub.src.utils.preset_utils import load_json

backbone_cls = SmolLM3Backbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "hidden_dim": transformers_config["hidden_size"],
        "num_layers": transformers_config["num_hidden_layers"],
        "num_attention_heads": transformers_config["num_attention_heads"],
        "num_key_value_heads": transformers_config["num_key_value_heads"],
        "intermediate_dim": transformers_config["intermediate_size"],
        "layer_norm_epsilon": transformers_config[
            "rms_norm_eps"
        ],  # Using rms_norm_eps as layer_norm_epsilon
        "max_position_embeddings": transformers_config[
            "max_position_embeddings"
        ],
        "rope_theta": transformers_config["rope_theta"],
        # partial_rotary_factor is not explicitly in config.json
        # but is inherited from the default value in the
        # `_compute_default_rope_parameters()` function
        "partial_rotary_factor": 1.0,
        "attention_bias": transformers_config["attention_bias"],
        "attention_dropout": transformers_config["attention_dropout"],
        # Despite the name, no_rope_layers: 1 = HAS RoPE, 0 = NO RoPE
        "rope_layer_enabled_list": [
            bool(x) for x in transformers_config["no_rope_layers"]
        ],
        "layer_types": transformers_config["layer_types"],
        "mlp_bias": transformers_config["mlp_bias"],
    }


def convert_weights(backbone, loader, transformers_config):
    loader.port_weight(
        keras_variable=backbone.get_layer("token_embedding").embeddings,
        hf_weight_key="model.embed_tokens.weight",
    )

    def transpose_and_reshape(x, shape):
        return np.reshape(np.transpose(x), shape)

    for i in range(backbone.num_layers):
        decoder_layer = backbone.get_layer(f"transformer_layer_{i}")

        # Input layernorm
        loader.port_weight(
            keras_variable=decoder_layer.input_layernorm.scale,
            hf_weight_key=f"model.layers.{i}.input_layernorm.weight",
        )

        # Attention layers
        ## Query
        loader.port_weight(
            keras_variable=decoder_layer.self_attn.q_proj.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.q_proj.weight",
            hook_fn=transpose_and_reshape,
        )
        ## Key
        loader.port_weight(
            keras_variable=decoder_layer.self_attn.k_proj.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.k_proj.weight",
            hook_fn=transpose_and_reshape,
        )
        ## Value
        loader.port_weight(
            keras_variable=decoder_layer.self_attn.v_proj.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.v_proj.weight",
            hook_fn=transpose_and_reshape,
        )
        ## Output
        loader.port_weight(
            keras_variable=decoder_layer.self_attn.o_proj.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.o_proj.weight",
            hook_fn=transpose_and_reshape,
        )

        # MLP layers
        loader.port_weight(
            keras_variable=decoder_layer.mlp.up_proj.kernel,
            hf_weight_key=f"model.layers.{i}.mlp.up_proj.weight",
            # rearrange_patterns="b a -> a b",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer.mlp.down_proj.kernel,
            hf_weight_key=f"model.layers.{i}.mlp.down_proj.weight",
            # rearrange_patterns="b a -> a b",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer.mlp.gate_proj.kernel,
            hf_weight_key=f"model.layers.{i}.mlp.gate_proj.weight",
            # rearrange_patterns="b a -> a b",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

        # Feedforward layernorm
        loader.port_weight(
            keras_variable=decoder_layer.post_attention_layernorm.scale,
            hf_weight_key=f"model.layers.{i}.post_attention_layernorm.weight",
        )

    # Final normalization layer
    loader.port_weight(
        keras_variable=backbone.get_layer("sequence_output_layernorm").scale,
        hf_weight_key="model.norm.weight",
    )

    backbone.training = False

    return backbone


def convert_tokenizer(cls, preset, **kwargs):
    tokenizer_config = load_json(preset, "tokenizer.json")
    vocab = tokenizer_config["model"]["vocab"]
    merges = tokenizer_config["model"]["merges"]
    merges = [" ".join(item) for item in merges]

    # Load all special tokens with the exception of "reserved" ones.
    special_tokens = set()
    for token in tokenizer_config["added_tokens"]:
        if not token["content"].startswith("<|reserved_special_token_"):
            vocab[token["content"]] = token["id"]
            special_tokens.add(token["content"])

    kwargs.update(
        {
            "unsplittable_tokens": list(special_tokens),
        }
    )

    return cls(vocabulary=vocab, merges=merges, **kwargs)



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.feature_pyramid_backbone import FeaturePyramidBackbone
from keras_hub.src.models.retinanet.feature_pyramid import FeaturePyramid
from keras_hub.src.utils.keras_utils import standardize_data_format


@keras_hub_export("keras_hub.models.RetinaNetBackbone")
class RetinaNetBackbone(FeaturePyramidBackbone):
    """RetinaNet Backbone.

    Combines a CNN backbone (e.g., ResNet, MobileNet) with a feature pyramid
    network (FPN)to extract multi-scale features for object detection.

    Args:
        image_encoder: `keras.Model`. The backbone model (e.g., ResNet50,
            MobileNetV2) used to extract features from the input image.
            It should have pyramid outputs (i.e., a dictionary mapping level
            names like `"P2"`, `"P3"`, etc. to their corresponding feature
            tensors).
        min_level: int. The minimum level of the feature pyramid (e.g., 3).
            This determines the coarsest level of features used.
        max_level: int. The maximum level of the feature pyramid (e.g., 7).
            This determines the finest level of features used.
        use_p5: bool. Determines the input source for creating coarser
            feature pyramid levels. If `True`, the output of the last backbone
            layer (typically `'P5'` in an FPN) is used as input to create
            higher-level feature maps (e.g., `'P6'`, `'P7'`) through
            additional convolutional layers. If `False`, the original `'P5'`
            feature map from the backbone is directly used as input for
            creating the coarser levels, bypassing any further processing of
            `'P5'` within the feature pyramid. Defaults to `False`.
        use_fpn_batch_norm: bool. Whether to use batch normalization in the
            feature pyramid network. Defaults to `False`.
        image_shape: tuple. tuple. The shape of the input image (H, W, C).
            The height and width can be `None` if they are variable.
        data_format: str. The data format of the input image
            (channels_first or channels_last).
        dtype: str. The data type of the input image.
        **kwargs: Additional keyword arguments passed to the base class.

    Raises:
        ValueError: If `min_level` is greater than `max_level`.
        ValueError: If `backbone_max_level` is less than 5 and `max_level` is
            greater than or equal to 5.
    """

    def __init__(
        self,
        image_encoder,
        min_level,
        max_level,
        use_p5,
        use_fpn_batch_norm=False,
        image_shape=(None, None, 3),
        data_format=None,
        dtype=None,
        **kwargs,
    ):
        # === Layers ===
        if min_level > max_level:
            raise ValueError(
                f"Minimum level ({min_level}) must be less than or equal to "
                f"maximum level ({max_level})."
            )

        data_format = standardize_data_format(data_format)
        input_levels = [
            int(level[1]) for level in image_encoder.pyramid_outputs
        ]
        backbone_max_level = min(max(input_levels), max_level)

        if backbone_max_level < 5 and max_level >= 5:
            raise ValueError(
                f"Backbone maximum level ({backbone_max_level}) is less than "
                f"the desired maximum level ({max_level}). "
                f"Please ensure that the backbone can generate features up to "
                f"the specified maximum level."
            )
        feature_extractor = keras.Model(
            inputs=image_encoder.inputs,
            outputs={
                f"P{level}": image_encoder.pyramid_outputs[f"P{level}"]
                for level in range(min_level, backbone_max_level + 1)
            },
            name="backbone",
        )

        feature_pyramid = FeaturePyramid(
            min_level=min_level,
            max_level=max_level,
            use_p5=use_p5,
            name="fpn",
            dtype=dtype,
            data_format=data_format,
            use_batch_norm=use_fpn_batch_norm,
        )

        # === Functional model ===
        image_input = keras.layers.Input(image_shape, name="inputs")
        feature_extractor_outputs = feature_extractor(image_input)
        feature_pyramid_outputs = feature_pyramid(feature_extractor_outputs)

        super().__init__(
            inputs=image_input,
            outputs=feature_pyramid_outputs,
            dtype=dtype,
            **kwargs,
        )

        # === config ===
        self.min_level = min_level
        self.max_level = max_level
        self.use_p5 = use_p5
        self.use_fpn_batch_norm = use_fpn_batch_norm
        self.image_encoder = image_encoder
        self.feature_pyramid = feature_pyramid
        self.image_shape = image_shape
        self.pyramid_outputs = feature_pyramid_outputs

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "image_encoder": keras.layers.serialize(self.image_encoder),
                "min_level": self.min_level,
                "max_level": self.max_level,
                "use_p5": self.use_p5,
                "use_fpn_batch_norm": self.use_fpn_batch_norm,
                "image_shape": self.image_shape,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        config.update(
            {
                "image_encoder": keras.layers.deserialize(
                    config["image_encoder"]
                ),
            }
        )

        return super().from_config(config)



from keras_hub.src.models.vit.vit_backbone import ViTBackbone
from keras_hub.src.models.vit.vit_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, ViTBackbone)



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.basnet.basnet_backbone import BASNetBackbone
from keras_hub.src.models.basnet.basnet_preprocessor import BASNetPreprocessor
from keras_hub.src.models.image_segmenter import ImageSegmenter


@keras_hub_export("keras_hub.models.BASNetImageSegmenter")
class BASNetImageSegmenter(ImageSegmenter):
    """BASNet image segmentation task.

    Args:
        backbone: A `keras_hub.models.BASNetBackbone` instance.
        preprocessor: `None`, a `keras_hub.models.Preprocessor` instance,
            a `keras.Layer` instance, or a callable. If `None` no preprocessing
            will be applied to the inputs.

    Example:
    ```python
    import keras_hub

    images = np.ones(shape=(1, 288, 288, 3))
    labels = np.zeros(shape=(1, 288, 288, 1))

    image_encoder = keras_hub.models.ResNetBackbone.from_preset(
        "resnet_18_imagenet",
        load_weights=False
    )
    backbone = keras_hub.models.BASNetBackbone(
        image_encoder,
        num_classes=1,
        image_shape=[288, 288, 3]
    )
    model = keras_hub.models.BASNetImageSegmenter(backbone)

    # Evaluate the model
    pred_labels = model(images)

    # Train the model
    model.compile(
        optimizer="adam",
        loss=keras.losses.BinaryCrossentropy(from_logits=False),
        metrics=["accuracy"],
    )
    model.fit(images, labels, epochs=3)
    ```
    """

    backbone_cls = BASNetBackbone
    preprocessor_cls = BASNetPreprocessor

    def __init__(
        self,
        backbone,
        preprocessor=None,
        **kwargs,
    ):
        # === Functional Model ===
        x = backbone.input
        outputs = backbone(x)
        # only return the refinement module's output as final prediction
        outputs = outputs["refine_out"]
        super().__init__(inputs=x, outputs=outputs, **kwargs)

        # === Config ===
        self.backbone = backbone
        self.preprocessor = preprocessor

    def compute_loss(self, x, y, y_pred, *args, **kwargs):
        # train BASNet's prediction and refinement module outputs against the
        # same ground truth data
        outputs = self.backbone(x)
        losses = []
        for output in outputs.values():
            losses.append(super().compute_loss(x, y, output, *args, **kwargs))
        return keras.ops.sum(losses, axis=0)

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        metrics="auto",
        **kwargs,
    ):
        """Configures the `BASNet` task for training.

        `BASNet` extends the default compilation signature
        of `keras.Model.compile` with defaults for `optimizer` and `loss`. To
        override these defaults, pass any value to these arguments during
        compilation.

        Args:
            optimizer: `"auto"`, an optimizer name, or a `keras.Optimizer`
                instance. Defaults to `"auto"`, which uses the default
                optimizer for `BASNet`. See `keras.Model.compile` and
                `keras.optimizers` for more info on possible `optimizer`
                values.
            loss: `"auto"`, a loss name, or a `keras.losses.Loss` instance.
                Defaults to `"auto"`, in which case the default loss
                computation of `BASNet` will be applied.
                See `keras.Model.compile` and `keras.losses` for more info on
                possible `loss` values.
            metrics: `"auto"`, or a list of metrics to be evaluated by
                the model during training and testing. Defaults to `"auto"`,
                where a `keras.metrics.Accuracy` will be applied to track the
                accuracy of the model during training.
                See `keras.Model.compile` and `keras.metrics` for
                more info on possible `metrics` values.
            **kwargs: See `keras.Model.compile` for a full list of arguments
                supported by the compile method.
        """
        if loss == "auto":
            loss = keras.losses.BinaryCrossentropy()
        if metrics == "auto":
            metrics = [keras.metrics.Accuracy()]
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            **kwargs,
        )



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.modeling.reversible_embedding import (
    ReversibleEmbedding,
)
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.moonshine.moonshine_decoder import (
    MoonshineDecoderBlock,
)
from keras_hub.src.models.moonshine.moonshine_encoder import (
    MoonshineEncoderBlock,
)
from keras_hub.src.models.moonshine.moonshine_layers import (
    MoonshineRotaryEmbedding,
)
from keras_hub.src.models.moonshine.moonshine_layers import (
    moonshine_kernel_initializer,
)


def compute_output_lengths(input_lengths):
    lengths = keras.ops.cast(input_lengths, "float32")
    lengths = keras.ops.floor((lengths - 127) / 64) + 1
    lengths = keras.ops.floor((lengths - 7) / 3) + 1
    lengths = keras.ops.floor((lengths - 3) / 2) + 1
    return keras.ops.maximum(keras.ops.cast(lengths, "int32"), 0)


@keras.saving.register_keras_serializable(package="keras_hub")
class ComputeAttentionMask(keras.layers.Layer):
    def call(self, features_for_shape, output_lengths):
        max_output_length = keras.ops.shape(features_for_shape)[1]
        indices = keras.ops.arange(max_output_length, dtype="int32")
        attention_mask = indices[None, :] < output_lengths[:, None]
        attention_mask = keras.ops.cast(attention_mask, "bool")
        return attention_mask

    def compute_output_shape(self, input_shapes):
        batch_dim = None
        if isinstance(input_shapes, (list, tuple)) and len(input_shapes) > 0:
            features_shape = input_shapes[0]
            if (
                isinstance(features_shape, (list, tuple))
                and len(features_shape) > 0
            ):
                batch_dim = features_shape[0]
        return (batch_dim, None)


@keras.saving.register_keras_serializable(package="keras_hub")
class Arange(keras.layers.Layer):
    def call(self, inputs):
        sequence_length = keras.ops.shape(inputs)[1]
        return keras.ops.arange(sequence_length, dtype="int32")


@keras_hub_export("keras_hub.models.MoonshineBackbone")
class MoonshineBackbone(Backbone):
    """Moonshine backbone with integrated audio feature extraction.

    This class implements an encoder-decoder backbone, as used in the Moonshine
    ASR system. It includes initial convolutional layers for audio feature
    extraction followed by `MoonshineEncoderBlock` instances for processing
    these features and `MoonshineDecoderBlock` instances for generating output
    sequences.

    Args:
        vocabulary_size: int. The size of the vocabulary for the embedding
            layers.
        filter_dim: int. The number of filters for the initial convolutional
            feature extractor layers. Typically matches `hidden_dim`.
        encoder_num_layers: int. The number of stacked encoder blocks.
        decoder_num_layers: int. The number of stacked decoder blocks.
        hidden_dim: int. The dimensionality of the model's hidden
            representations and embeddings.
        intermediate_dim: int. The dimensionality of the intermediate
            representations in feedforward networks.
        encoder_num_heads: int. The number of attention heads in the encoder's
            multi-head attention.
        decoder_num_heads: int. The number of attention heads in the decoder's
            multi-head attention.
        feedforward_expansion_factor: int, optional. A multiplier applied to
            `intermediate_dim` to determine the total width of the feedforward
            network. Defaults to 4.
        encoder_use_swiglu_activation: bool, optional. When True, uses SwiGLU
            in the encoder feedforward network. Defaults to False.
        decoder_use_swiglu_activation: bool, optional. When True, uses SwiGLU
            in the decoder feedforward network. Defaults to True.
        max_position_embeddings: int, optional. The maximum sequence length for
            position embeddings. Defaults to 2048.
        pad_head_dim_to_multiple_of: int, optional. If specified, pads the head
            dimension to be a multiple of this value for performance
            optimization. Defaults to None.
        partial_rotary_factor: float, optional. The fraction of dimensions to
            apply rotary position embeddings to. Defaults to 0.62.
        dropout: float, optional. The dropout probability for input dropout
            layers. Defaults to 0.0.
        initializer_range: float, optional. The standard deviation of the
            truncated normal initializer for weights. Defaults to 0.02.
        rope_theta: float, optional. The base frequency for rotary position
            embeddings. Defaults to 10,000.0.
        attention_bias: bool, optional. Whether to use bias in attention
            mechanisms. Defaults to False.
        attention_dropout: float, optional. The dropout probability for
            attention mechanisms. Defaults to 0.0.
        dtype: str, optional. The dtype to use for model computations and
            weights. Defaults to None.

    Examples:
    ```python
    import numpy as np
    import keras
    from keras_hub.models import MoonshineBackbone

    # Create random input data for demonstration.
    # Input is now raw-ish audio features (e.g., from MoonshineAudioConverter).
    encoder_raw_input_values = np.random.rand(1, 16000, 1).astype("float32")
    # Mask corresponding to the raw input time dimension
    encoder_padding_mask = np.ones((1, 16000), dtype="bool")
    decoder_token_ids = np.random.randint(
        0, 1000, size=(1, 20), dtype="int32"
    )
    decoder_padding_mask = np.ones((1, 20), dtype="bool")

    # Initialize the Moonshine backbone with specific parameters.
    backbone = MoonshineBackbone(
        vocabulary_size=10000,
        filter_dim=256,
        encoder_num_layers=6,
        decoder_num_layers=6,
        hidden_dim=256,
        intermediate_dim=512,
        encoder_num_heads=8,
        decoder_num_heads=8,
        feedforward_expansion_factor=4,
        decoder_use_swiglu_activation=True,
        encoder_use_swiglu_activation=False,
    )

    # Forward pass through the model.
    outputs = backbone(
        {
            "encoder_input_values": encoder_raw_input_values,
            "encoder_padding_mask": encoder_padding_mask,
            "decoder_token_ids": decoder_token_ids,
            "decoder_padding_mask": decoder_padding_mask,
        }
    )

    # Display the outputs.
    print("Encoder output shape:", outputs["encoder_sequence_output"].shape)
    print("Decoder output shape:", outputs["decoder_sequence_output"].shape)
    ```
    """

    # References:
    # Feature Extractor: UsefulSensors implementation (https://github.com/usefulsensors/moonshine/blob/4a000427bd36a1c2c6d20a86c672dbd850b44c88/moonshine/model.py#L6-L32)
    # Transformer Backbone: Hugging Face implementation (https://github.com/huggingface/transformers/blob/dcbdf7e962c4b36140cc9ee76f870016121e69e5/src/transformers/models/moonshine/modeling_moonshine.py#L1326-L1486).

    def __init__(
        self,
        vocabulary_size,
        filter_dim,
        encoder_num_layers,
        decoder_num_layers,
        hidden_dim,
        intermediate_dim,
        encoder_num_heads,
        decoder_num_heads,
        feedforward_expansion_factor=4,
        encoder_use_swiglu_activation=False,
        decoder_use_swiglu_activation=True,
        max_position_embeddings=2048,
        pad_head_dim_to_multiple_of=None,
        partial_rotary_factor=0.62,
        dropout=0.0,
        initializer_range=0.02,
        rope_theta=10000.0,
        attention_bias=False,
        attention_dropout=0.0,
        dtype=None,
        **kwargs,
    ):
        # ==== Layers ====
        self._compute_mask_layer = ComputeAttentionMask(
            name="compute_attention_mask"
        )

        # Feature extractor layers.
        self.conv1 = keras.layers.Conv1D(
            filters=filter_dim,
            kernel_size=127,
            strides=64,
            use_bias=False,
            padding="valid",
            kernel_initializer=moonshine_kernel_initializer(
                initializer_range=initializer_range
            ),
            name="conv1",
            dtype=dtype,
        )
        self.group_norm = keras.layers.GroupNormalization(
            groups=1,
            axis=-1,
            epsilon=1e-5,
            center=True,
            scale=True,
            name="group_norm",
            dtype=dtype,
        )
        self.tanh_after_conv1 = keras.layers.Activation(
            "tanh", name="tanh_after_conv1", dtype=dtype
        )
        self.conv2 = keras.layers.Conv1D(
            filters=2 * filter_dim,
            kernel_size=7,
            strides=3,
            use_bias=True,
            padding="valid",
            kernel_initializer=moonshine_kernel_initializer(
                initializer_range=initializer_range
            ),
            name="conv2",
            dtype=dtype,
        )
        self.gelu_after_conv2 = keras.layers.Activation(
            "gelu", name="gelu_after_conv2", dtype=dtype
        )
        self.conv3 = keras.layers.Conv1D(
            filters=filter_dim,
            kernel_size=3,
            strides=2,
            use_bias=True,
            padding="valid",
            kernel_initializer=moonshine_kernel_initializer(
                initializer_range=initializer_range
            ),
            name="conv3",
            dtype=dtype,
        )
        self.gelu_after_conv3 = keras.layers.Activation(
            "gelu", name="gelu_after_conv3", dtype=dtype
        )

        # Transformer layers.
        encoder_head_dim = hidden_dim // encoder_num_heads
        if pad_head_dim_to_multiple_of:
            encoder_head_dim = (
                (encoder_head_dim + pad_head_dim_to_multiple_of - 1)
                // pad_head_dim_to_multiple_of
            ) * pad_head_dim_to_multiple_of

        decoder_head_dim = hidden_dim // decoder_num_heads
        if pad_head_dim_to_multiple_of:
            decoder_head_dim = (
                (decoder_head_dim + pad_head_dim_to_multiple_of - 1)
                // pad_head_dim_to_multiple_of
            ) * pad_head_dim_to_multiple_of

        # Embedding layer for decoder.
        self.token_embedding = ReversibleEmbedding(
            input_dim=vocabulary_size,
            output_dim=hidden_dim,
            embeddings_initializer=moonshine_kernel_initializer(
                initializer_range=initializer_range
            ),
            name="token_embedding",
            dtype=dtype,
        )

        # Rotary embeddings for encoder and decoder.
        self.encoder_rotary_embedding = MoonshineRotaryEmbedding(
            head_dim=encoder_head_dim,
            max_position_embeddings=max_position_embeddings,
            partial_rotary_factor=partial_rotary_factor,
            base_value=rope_theta,
            name="encoder_rotary_embedding",
            dtype=dtype,
        )

        self.decoder_rotary_embedding = MoonshineRotaryEmbedding(
            head_dim=decoder_head_dim,
            max_position_embeddings=max_position_embeddings,
            partial_rotary_factor=partial_rotary_factor,
            base_value=rope_theta,
            name="decoder_rotary_embedding",
            dtype=dtype,
        )

        # Dropout for encoder.
        self.encoder_dropout = keras.layers.Dropout(
            dropout, name="encoder_dropout", dtype=dtype
        )
        # Dropout for decoder.
        self.decoder_dropout = keras.layers.Dropout(
            dropout, name="decoder_dropout", dtype=dtype
        )

        # Encoder blocks.
        self.encoder_blocks = []
        for i in range(encoder_num_layers):
            encoder_block = MoonshineEncoderBlock(
                hidden_dim=hidden_dim,
                intermediate_dim=intermediate_dim,
                num_heads=encoder_num_heads,
                feedforward_expansion_factor=feedforward_expansion_factor,
                use_swiglu_activation=encoder_use_swiglu_activation,
                pad_head_dim_to_multiple_of=pad_head_dim_to_multiple_of,
                initializer_range=initializer_range,
                attention_bias=attention_bias,
                attention_dropout=attention_dropout,
                name=f"encoder_block_{i}",
                dtype=dtype,
            )
            self.encoder_blocks.append(encoder_block)

        # Layer normalization for encoder.
        self.encoder_final_layer_norm = keras.layers.LayerNormalization(
            epsilon=1e-5,
            center=False,
            scale=True,
            name="encoder_final_layer_norm",
            dtype=dtype,
        )

        # Decoder blocks.
        self.decoder_blocks = []
        for i in range(decoder_num_layers):
            decoder_block = MoonshineDecoderBlock(
                hidden_dim=hidden_dim,
                intermediate_dim=intermediate_dim,
                num_heads=decoder_num_heads,
                feedforward_expansion_factor=feedforward_expansion_factor,
                use_swiglu_activation=decoder_use_swiglu_activation,
                pad_head_dim_to_multiple_of=pad_head_dim_to_multiple_of,
                initializer_range=initializer_range,
                attention_bias=attention_bias,
                attention_dropout=attention_dropout,
                name=f"decoder_block_{i}",
                dtype=dtype,
            )
            self.decoder_blocks.append(decoder_block)

        # Layer normalization for decoder.
        self.decoder_post_norm = keras.layers.LayerNormalization(
            epsilon=1e-5,
            center=False,
            scale=True,
            name="decoder_post_norm",
            dtype=dtype,
        )

        # === Functional Model ===
        encoder_raw_input_values = keras.Input(
            shape=(None, 1), name="encoder_input_values", dtype=dtype
        )
        encoder_input_padding_mask = keras.Input(
            shape=(None,), name="encoder_padding_mask", dtype="bool"
        )
        decoder_input = keras.Input(
            shape=(None,), name="decoder_token_ids", dtype="int32"
        )
        decoder_padding_mask = keras.Input(
            shape=(None,), name="decoder_padding_mask", dtype="bool"
        )

        # Feature extraction.
        encoder_hidden_states = self.conv1(encoder_raw_input_values)
        encoder_hidden_states = self.tanh_after_conv1(encoder_hidden_states)
        encoder_hidden_states = self.group_norm(encoder_hidden_states)
        encoder_hidden_states = self.conv2(encoder_hidden_states)
        encoder_hidden_states = self.gelu_after_conv2(encoder_hidden_states)
        encoder_hidden_states = self.conv3(encoder_hidden_states)
        encoder_hidden_states = self.gelu_after_conv3(encoder_hidden_states)

        # Compute mask for encoder features.
        original_lengths = keras.ops.sum(
            keras.ops.cast(encoder_input_padding_mask, "int32"), axis=1
        )
        output_lengths = compute_output_lengths(original_lengths)
        encoder_attention_mask = self._compute_mask_layer(
            encoder_hidden_states, output_lengths
        )

        # Encoder.
        encoder_positions = Arange(name="encoder_positions")(
            encoder_hidden_states
        )
        encoder_rotary_emb = self.encoder_rotary_embedding(encoder_positions)
        encoder_hidden_states = self.encoder_dropout(encoder_hidden_states)
        for encoder_block in self.encoder_blocks:
            encoder_hidden_states = encoder_block(
                encoder_hidden_states,
                encoder_rotary_emb,
                attention_mask=encoder_attention_mask,
            )
        encoder_output = self.encoder_final_layer_norm(encoder_hidden_states)

        # Decoder.
        decoder_positions = Arange(name="decoder_positions")(decoder_input)
        decoder_rotary_emb = self.decoder_rotary_embedding(decoder_positions)
        decoder_hidden_states = self.token_embedding(decoder_input)
        decoder_hidden_states = self.decoder_dropout(decoder_hidden_states)
        for decoder_block in self.decoder_blocks:
            decoder_hidden_states = decoder_block(
                decoder_sequence=decoder_hidden_states,
                encoder_sequence=encoder_output,
                rotary_embedding=decoder_rotary_emb,
                decoder_padding_mask=decoder_padding_mask,
                encoder_padding_mask=encoder_attention_mask,
            )
        decoder_output = self.decoder_post_norm(decoder_hidden_states)

        super().__init__(
            inputs={
                "encoder_input_values": encoder_raw_input_values,
                "encoder_padding_mask": encoder_input_padding_mask,
                "decoder_token_ids": decoder_input,
                "decoder_padding_mask": decoder_padding_mask,
            },
            outputs={
                "encoder_sequence_output": encoder_output,
                "decoder_sequence_output": decoder_output,
                "encoder_attention_mask": encoder_attention_mask,
            },
            dtype=dtype,
            **kwargs,
        )

        # ==== Config ====
        self.vocabulary_size = vocabulary_size
        self.filter_dim = filter_dim
        self.encoder_num_layers = encoder_num_layers
        self.decoder_num_layers = decoder_num_layers
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.encoder_num_heads = encoder_num_heads
        self.decoder_num_heads = decoder_num_heads
        self.feedforward_expansion_factor = feedforward_expansion_factor
        self.encoder_use_swiglu_activation = encoder_use_swiglu_activation
        self.decoder_use_swiglu_activation = decoder_use_swiglu_activation
        self.max_position_embeddings = max_position_embeddings
        self.pad_head_dim_to_multiple_of = pad_head_dim_to_multiple_of
        self.partial_rotary_factor = partial_rotary_factor
        self.dropout = dropout
        self.initializer_range = initializer_range
        self.rope_theta = rope_theta
        self.attention_bias = attention_bias
        self.attention_dropout = attention_dropout

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary_size": self.vocabulary_size,
                "filter_dim": self.filter_dim,
                "encoder_num_layers": self.encoder_num_layers,
                "decoder_num_layers": self.decoder_num_layers,
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "encoder_num_heads": self.encoder_num_heads,
                "decoder_num_heads": self.decoder_num_heads,
                "feedforward_expansion_factor": self.feedforward_expansion_factor,  # noqa: E501
                "encoder_use_swiglu_activation": self.encoder_use_swiglu_activation,  # noqa: E501
                "decoder_use_swiglu_activation": self.decoder_use_swiglu_activation,  # noqa: E501
                "max_position_embeddings": self.max_position_embeddings,
                "pad_head_dim_to_multiple_of": self.pad_head_dim_to_multiple_of,
                "partial_rotary_factor": self.partial_rotary_factor,
                "dropout": self.dropout,
                "initializer_range": self.initializer_range,
                "rope_theta": self.rope_theta,
                "attention_bias": self.attention_bias,
                "attention_dropout": self.attention_dropout,
                "dtype": self.dtype,
            }
        )
        return config



from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone
from keras_hub.src.models.hgnetv2.hgnetv2_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, HGNetV2Backbone)



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.modeling.non_max_supression import NonMaxSuppression
from keras_hub.src.models.d_fine.d_fine_backbone import DFineBackbone
from keras_hub.src.models.d_fine.d_fine_loss import compute_box_losses
from keras_hub.src.models.d_fine.d_fine_loss import compute_local_losses
from keras_hub.src.models.d_fine.d_fine_loss import compute_vfl_loss
from keras_hub.src.models.d_fine.d_fine_loss import get_cdn_matched_indices
from keras_hub.src.models.d_fine.d_fine_loss import hungarian_matcher
from keras_hub.src.models.d_fine.d_fine_object_detector_preprocessor import (
    DFineObjectDetectorPreprocessor,
)
from keras_hub.src.models.object_detector import ObjectDetector
from keras_hub.src.utils.tensor_utils import assert_bounding_box_support


@keras_hub_export("keras_hub.models.DFineObjectDetector")
class DFineObjectDetector(ObjectDetector):
    """D-FINE Object Detector model.

    This class wraps the `DFineBackbone` and adds the final prediction and loss
    computation logic for end-to-end object detection. It is responsible for:
    1.  Defining the functional model that connects the `DFineBackbone` to the
        input layers.
    2.  Implementing the `compute_loss` method, which uses a Hungarian matcher
        to assign predictions to ground truth targets and calculates a weighted
        sum of multiple loss components (classification, bounding box, etc.).
    3.  Post-processing the raw outputs from the backbone into final, decoded
        predictions (boxes, labels, confidence scores) during inference.

    Args:
        backbone: A `keras_hub.models.Backbone` instance, specifically a
            `DFineBackbone`, serving as the feature extractor for the object
            detector.
        num_classes: An integer representing the number of object classes to
            detect.
        bounding_box_format: A string specifying the format of the bounding
            boxes. Defaults to `"yxyx"`. Must be a supported format (e.g.,
            `"yxyx"`, `"xyxy"`).
        preprocessor: Optional. An instance of `DFineObjectDetectorPreprocessor`
            for input data preprocessing.
        matcher_class_cost: A float representing the cost for class mismatch in
            the Hungarian matcher. Defaults to `2.0`.
        matcher_bbox_cost: A float representing the cost for bounding box
            mismatch in the Hungarian matcher. Defaults to `5.0`.
        matcher_ciou_cost: A float representing the cost for complete IoU
            mismatch in the Hungarian matcher. Defaults to `2.0`.
        use_focal_loss: A boolean indicating whether to use focal loss for
            classification. Defaults to `True`.
        matcher_alpha: A float parameter for the focal loss alpha. Defaults to
            `0.25`.
        matcher_gamma: A float parameter for the focal loss gamma. Defaults to
            `2.0`.
        weight_loss_vfl: Weight for the classification loss. Defaults to `1.0`.
        weight_loss_bbox: Weight for the bounding box regression loss. Default
            is `5.0`.
        weight_loss_ciou: Weight for the complete IoU loss. Defaults to
            `2.0`.
        weight_loss_fgl: Weight for the focal grid loss. Defaults to `0.15`.
        weight_loss_ddf: Weight for the DDF loss. Defaults to `1.5`.
        ddf_temperature: A float temperature scaling factor for the DDF loss.
            Defaults to `5.0`.
        prediction_decoder: Optional. A `keras.layers.Layer` instance that
            decodes raw predictions. If not provided, a `NonMaxSuppression`
            layer is used.
        activation: Optional. The activation function to apply to the logits
            before decoding. Defaults to `None`.

    Examples:

    **Creating a DFineObjectDetector without labels:**

    ```python
    import numpy as np
    from keras_hub.src.models.d_fine.d_fine_object_detector import (
        DFineObjectDetector
    )
    from keras_hub.src.models.d_fine.d_fine_backbone import DFineBackbone
    from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone

    # Initialize the backbone without labels.
    hgnetv2_backbone = HGNetV2Backbone(
        stem_channels=[3, 16, 16],
        stackwise_stage_filters=[
            [16, 16, 64, 1, 3, 3],
            [64, 32, 256, 1, 3, 3],
            [256, 64, 512, 2, 3, 5],
            [512, 128, 1024, 1, 3, 5],
        ],
        apply_downsample=[False, True, True, True],
        use_lightweight_conv_block=[False, False, True, True],
        depths=[1, 1, 2, 1],
        hidden_sizes=[64, 256, 512, 1024],
        embedding_size=16,
        use_learnable_affine_block=True,
        hidden_act="relu",
        image_shape=(256, 256, 3),
        out_features=["stage3", "stage4"],
    )

    # Initialize the backbone without labels.
    backbone = DFineBackbone(
        backbone=hgnetv2_backbone,
        decoder_in_channels=[128, 128],
        encoder_hidden_dim=128,
        num_denoising=100,
        num_labels=80,
        hidden_dim=128,
        learn_initial_query=False,
        num_queries=300,
        anchor_image_size=(256, 256),
        feat_strides=[16, 32],
        num_feature_levels=2,
        encoder_in_channels=[512, 1024],
        encode_proj_layers=[1],
        num_attention_heads=8,
        encoder_ffn_dim=512,
        num_encoder_layers=1,
        hidden_expansion=0.34,
        depth_multiplier=0.5,
        eval_idx=-1,
        num_decoder_layers=3,
        decoder_attention_heads=8,
        decoder_ffn_dim=512,
        decoder_n_points=[6, 6],
        lqe_hidden_dim=64,
        num_lqe_layers=2,
        out_features=["stage3", "stage4"],
        image_shape=(256, 256, 3),
    )

    # Create the detector.
    detector = DFineObjectDetector(
        backbone=backbone,
        num_classes=80,
        bounding_box_format="yxyx",
    )
    ```

    **Creating a DFineObjectDetector with labels for the backbone:**

    ```python
    import numpy as np
    from keras_hub.src.models.d_fine.d_fine_object_detector import (
        DFineObjectDetector
    )
    from keras_hub.src.models.d_fine.d_fine_backbone import DFineBackbone
    from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone

    # Define labels for the backbone.
    labels = [
        {
            "boxes": np.array([[0.5, 0.5, 0.2, 0.2], [0.4, 0.4, 0.1, 0.1]]),
            "labels": np.array([1, 10])
        },
        {"boxes": np.array([[0.6, 0.6, 0.3, 0.3]]), "labels": np.array([20])},
    ]

    hgnetv2_backbone = HGNetV2Backbone(
        stem_channels=[3, 16, 16],
        stackwise_stage_filters=[
            [16, 16, 64, 1, 3, 3],
            [64, 32, 256, 1, 3, 3],
            [256, 64, 512, 2, 3, 5],
            [512, 128, 1024, 1, 3, 5],
        ],
        apply_downsample=[False, True, True, True],
        use_lightweight_conv_block=[False, False, True, True],
        depths=[1, 1, 2, 1],
        hidden_sizes=[64, 256, 512, 1024],
        embedding_size=16,
        use_learnable_affine_block=True,
        hidden_act="relu",
        image_shape=(256, 256, 3),
        out_features=["stage3", "stage4"],
    )

    # Backbone is initialized with labels.
    backbone = DFineBackbone(
        backbone=hgnetv2_backbone,
        decoder_in_channels=[128, 128],
        encoder_hidden_dim=128,
        num_denoising=100,
        num_labels=80,
        hidden_dim=128,
        learn_initial_query=False,
        num_queries=300,
        anchor_image_size=(256, 256),
        feat_strides=[16, 32],
        num_feature_levels=2,
        encoder_in_channels=[512, 1024],
        encode_proj_layers=[1],
        num_attention_heads=8,
        encoder_ffn_dim=512,
        num_encoder_layers=1,
        hidden_expansion=0.34,
        depth_multiplier=0.5,
        eval_idx=-1,
        num_decoder_layers=3,
        decoder_attention_heads=8,
        decoder_ffn_dim=512,
        decoder_n_points=[6, 6],
        lqe_hidden_dim=64,
        num_lqe_layers=2,
        out_features=["stage3", "stage4"],
        image_shape=(256, 256, 3),
        labels=labels,
        box_noise_scale=1.0,
        label_noise_ratio=0.5,
    )

    # Create the detector.
    detector = DFineObjectDetector(
        backbone=backbone,
        num_classes=80,
        bounding_box_format="yxyx",
    )
    ```

    **Using the detector for training:**

    ```python
    import numpy as np
    from keras_hub.src.models.d_fine.d_fine_object_detector import (
        DFineObjectDetector
    )
    from keras_hub.src.models.d_fine.d_fine_backbone import DFineBackbone
    from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone

    # Initialize backbone and detector.
    hgnetv2_backbone = HGNetV2Backbone(
        stem_channels=[3, 16, 16],
        stackwise_stage_filters=[
            [16, 16, 64, 1, 3, 3],
            [64, 32, 256, 1, 3, 3],
            [256, 64, 512, 2, 3, 5],
            [512, 128, 1024, 1, 3, 5],
        ],
        apply_downsample=[False, True, True, True],
        use_lightweight_conv_block=[False, False, True, True],
        depths=[1, 1, 2, 1],
        hidden_sizes=[64, 256, 512, 1024],
        embedding_size=16,
        use_learnable_affine_block=True,
        hidden_act="relu",
        image_shape=(256, 256, 3),
        out_features=["stage3", "stage4"],
    )
    backbone = DFineBackbone(
        backbone=hgnetv2_backbone,
        decoder_in_channels=[128, 128],
        encoder_hidden_dim=128,
        num_denoising=100,
        num_labels=80,
        hidden_dim=128,
        learn_initial_query=False,
        num_queries=300,
        anchor_image_size=(256, 256),
        feat_strides=[16, 32],
        num_feature_levels=2,
        encoder_in_channels=[512, 1024],
        encode_proj_layers=[1],
        num_attention_heads=8,
        encoder_ffn_dim=512,
        num_encoder_layers=1,
        hidden_expansion=0.34,
        depth_multiplier=0.5,
        eval_idx=-1,
        num_decoder_layers=3,
        decoder_attention_heads=8,
        decoder_ffn_dim=512,
        decoder_n_points=[6, 6],
        lqe_hidden_dim=64,
        num_lqe_layers=2,
        out_features=["stage3", "stage4"],
        image_shape=(256, 256, 3),
    )
    detector = DFineObjectDetector(
        backbone=backbone,
        num_classes=80,
        bounding_box_format="yxyx",
    )

    # Sample training data.
    images = np.random.uniform(
        low=0, high=255, size=(2, 256, 256, 3)
    ).astype("float32")
    bounding_boxes = {
        "boxes": [
            np.array([[10.0, 20.0, 20.0, 30.0], [20.0, 30.0, 30.0, 40.0]]),
            np.array([[15.0, 25.0, 25.0, 35.0]]),
        ],
        "labels": [
            np.array([0, 2]), np.array([1])
        ],
    }

    # Compile the model.
    detector.compile(
        optimizer="adam",
        loss=detector.compute_loss,
    )

    # Train the model.
    detector.fit(x=images, y=bounding_boxes, epochs=1, batch_size=1)
    ```

    **Making predictions:**

    ```python
    import numpy as np
    from keras_hub.src.models.d_fine.d_fine_object_detector import (
        DFineObjectDetector
    )
    from keras_hub.src.models.d_fine.d_fine_backbone import DFineBackbone
    from keras_hub.src.models.hgnetv2.hgnetv2_backbone import HGNetV2Backbone

    # Initialize backbone and detector.
    hgnetv2_backbone = HGNetV2Backbone(
        stem_channels=[3, 16, 16],
        stackwise_stage_filters=[
            [16, 16, 64, 1, 3, 3],
            [64, 32, 256, 1, 3, 3],
            [256, 64, 512, 2, 3, 5],
            [512, 128, 1024, 1, 3, 5],
        ],
        apply_downsample=[False, True, True, True],
        use_lightweight_conv_block=[False, False, True, True],
        depths=[1, 1, 2, 1],
        hidden_sizes=[64, 256, 512, 1024],
        embedding_size=16,
        use_learnable_affine_block=True,
        hidden_act="relu",
        image_shape=(256, 256, 3),
        out_features=["stage3", "stage4"],
    )
    backbone = DFineBackbone(
        backbone=hgnetv2_backbone,
        decoder_in_channels=[128, 128],
        encoder_hidden_dim=128,
        num_denoising=100,
        num_labels=80,
        hidden_dim=128,
        learn_initial_query=False,
        num_queries=300,
        anchor_image_size=(256, 256),
        feat_strides=[16, 32],
        num_feature_levels=2,
        encoder_in_channels=[512, 1024],
        encode_proj_layers=[1],
        num_attention_heads=8,
        encoder_ffn_dim=512,
        num_encoder_layers=1,
        hidden_expansion=0.34,
        depth_multiplier=0.5,
        eval_idx=-1,
        num_decoder_layers=3,
        decoder_attention_heads=8,
        decoder_ffn_dim=512,
        decoder_n_points=[6, 6],
        lqe_hidden_dim=64,
        num_lqe_layers=2,
        out_features=["stage3", "stage4"],
        image_shape=(256, 256, 3),
    )
    detector = DFineObjectDetector(
        backbone=backbone,
        num_classes=80,
        bounding_box_format="yxyx",
    )

    # Sample test image.
    test_image = np.random.uniform(
        low=0, high=255, size=(1, 256, 256, 3)
    ).astype("float32")

    # Make predictions.
    predictions = detector.predict(test_image)

    # Access predictions.
    boxes = predictions["boxes"]                    # Shape: (1, 100, 4)
    labels = predictions["labels"]                  # Shape: (1, 100)
    confidence = predictions["confidence"]          # Shape: (1, 100)
    num_detections = predictions["num_detections"]  # Shape: (1,)
    ```
    """

    backbone_cls = DFineBackbone
    preprocessor_cls = DFineObjectDetectorPreprocessor

    def __init__(
        self,
        backbone,
        num_classes,
        bounding_box_format="yxyx",
        preprocessor=None,
        matcher_class_cost=2.0,
        matcher_bbox_cost=5.0,
        matcher_ciou_cost=2.0,
        use_focal_loss=True,
        matcher_alpha=0.25,
        matcher_gamma=2.0,
        weight_loss_vfl=1.0,
        weight_loss_bbox=5.0,
        weight_loss_ciou=2.0,
        weight_loss_fgl=0.15,
        weight_loss_ddf=1.5,
        ddf_temperature=5.0,
        prediction_decoder=None,
        activation=None,
        **kwargs,
    ):
        assert_bounding_box_support(self.__class__.__name__)

        # === Functional Model ===
        image_input = keras.layers.Input(
            shape=backbone.image_shape, name="images"
        )
        outputs = backbone(image_input)
        intermediate_logits = outputs["intermediate_logits"]
        intermediate_reference_points = outputs["intermediate_reference_points"]
        intermediate_predicted_corners = outputs[
            "intermediate_predicted_corners"
        ]
        initial_reference_points = outputs["initial_reference_points"]
        logits = intermediate_logits[:, -1, :, :]
        pred_boxes = intermediate_reference_points[:, -1, :, :]
        model_outputs = {
            "logits": logits,
            "pred_boxes": pred_boxes,
            "intermediate_logits": intermediate_logits,
            "intermediate_reference_points": intermediate_reference_points,
            "intermediate_predicted_corners": intermediate_predicted_corners,
            "initial_reference_points": initial_reference_points,
            "enc_topk_logits": outputs["enc_topk_logits"],
            "enc_topk_bboxes": outputs["enc_topk_bboxes"],
        }
        if "dn_num_group" in outputs:
            model_outputs["dn_positive_idx"] = outputs["dn_positive_idx"]
            model_outputs["dn_num_group"] = outputs["dn_num_group"]
            model_outputs["dn_num_split"] = outputs["dn_num_split"]
        super().__init__(
            inputs=image_input,
            outputs=model_outputs,
            **kwargs,
        )

        # === Config ===
        self.backbone = backbone
        self.num_classes = num_classes
        self.bounding_box_format = bounding_box_format
        self.preprocessor = preprocessor
        self.matcher_class_cost = matcher_class_cost
        self.matcher_bbox_cost = matcher_bbox_cost
        self.matcher_ciou_cost = matcher_ciou_cost
        self.use_focal_loss = use_focal_loss
        self.matcher_alpha = matcher_alpha
        self.matcher_gamma = matcher_gamma
        self.weight_dict = {
            "loss_vfl": weight_loss_vfl,
            "loss_bbox": weight_loss_bbox,
            "loss_ciou": weight_loss_ciou,
            "loss_fgl": weight_loss_fgl,
            "loss_ddf": weight_loss_ddf,
        }
        self.ddf_temperature = ddf_temperature
        self.activation = activation
        self._prediction_decoder = prediction_decoder or NonMaxSuppression(
            from_logits=(self.activation != keras.activations.sigmoid),
            bounding_box_format=self.bounding_box_format,
            max_detections=backbone.num_queries,
        )

    def compute_loss(self, x, y, y_pred, sample_weight, **kwargs):
        gt_boxes = y["boxes"]
        gt_labels = y["labels"]
        batch_size = keras.ops.shape(gt_labels)[0]
        num_objects = keras.ops.shape(gt_labels)[1]
        num_targets_per_image = keras.ops.tile(
            keras.ops.expand_dims(num_objects, 0), [batch_size]
        )
        labels_for_item = keras.ops.reshape(gt_labels, [-1])
        boxes_for_item = keras.ops.reshape(gt_boxes, [-1, 4])
        targets = {"labels": labels_for_item, "boxes": boxes_for_item}

        intermediate_logits_all = y_pred["intermediate_logits"]
        intermediate_ref_points_all = y_pred["intermediate_reference_points"]
        predicted_corners_all = y_pred["intermediate_predicted_corners"]
        initial_ref_points_all = y_pred["initial_reference_points"]
        enc_topk_logits = y_pred["enc_topk_logits"]
        enc_topk_bboxes = y_pred["enc_topk_bboxes"]
        if "dn_num_group" in y_pred:
            denoising_meta_values = {
                "dn_positive_idx": y_pred["dn_positive_idx"],
                "dn_num_group": y_pred["dn_num_group"],
                "dn_num_split": y_pred["dn_num_split"],
            }
            dn_split_point = self.backbone.dn_split_point
            (
                dn_intermediate_logits,
                matching_intermediate_logits,
            ) = keras.ops.split(
                intermediate_logits_all, [dn_split_point], axis=2
            )
            (
                dn_intermediate_ref_points,
                matching_intermediate_ref_points,
            ) = keras.ops.split(
                intermediate_ref_points_all, [dn_split_point], axis=2
            )
            (
                dn_predicted_corners,
                matching_predicted_corners,
            ) = keras.ops.split(predicted_corners_all, [dn_split_point], axis=2)
            (
                dn_initial_ref_points,
                matching_initial_ref_points,
            ) = keras.ops.split(
                initial_ref_points_all, [dn_split_point], axis=2
            )
        else:
            denoising_meta_values = None
            matching_intermediate_logits = intermediate_logits_all
            matching_intermediate_ref_points = intermediate_ref_points_all
            matching_predicted_corners = predicted_corners_all
            matching_initial_ref_points = initial_ref_points_all
        matching_logits = matching_intermediate_logits[:, -1, :, :]
        matching_pred_boxes = matching_intermediate_ref_points[:, -1, :, :]
        outputs_without_aux = {
            "logits": matching_logits,
            "pred_boxes": keras.ops.clip(matching_pred_boxes, 0, 1),
        }
        indices = hungarian_matcher(
            outputs_without_aux,
            [targets],
            num_targets_per_image,
            self.use_focal_loss,
            self.matcher_alpha,
            self.matcher_gamma,
            self.matcher_bbox_cost,
            self.matcher_class_cost,
            self.matcher_ciou_cost,
            self.backbone,
        )
        num_boxes = keras.ops.shape(labels_for_item)[0]
        num_boxes = keras.ops.convert_to_tensor(num_boxes, dtype="float32")
        num_boxes = keras.ops.maximum(num_boxes, 1.0)
        losses = {}
        vfl_loss = compute_vfl_loss(
            outputs_without_aux,
            [targets],
            indices,
            num_boxes,
            self.num_classes,
            self.matcher_alpha,
            self.matcher_gamma,
        )
        losses.update(
            {
                k: v * self.weight_dict[k]
                for k, v in vfl_loss.items()
                if k in self.weight_dict
            }
        )
        box_losses = compute_box_losses(
            outputs_without_aux, [targets], indices, num_boxes
        )
        losses.update(
            {
                k: v * self.weight_dict[k]
                for k, v in box_losses.items()
                if k in self.weight_dict
            }
        )
        local_losses = compute_local_losses(
            {
                **outputs_without_aux,
                "pred_corners": matching_predicted_corners[:, -1, :, :],
                "ref_points": matching_initial_ref_points[:, -1, :, :],
                "teacher_corners": keras.ops.zeros_like(
                    matching_predicted_corners[:, -1, :, :]
                ),
                "teacher_logits": keras.ops.zeros_like(matching_logits),
            },
            [targets],
            indices,
            num_boxes,
            self.backbone,
            self.ddf_temperature,
            compute_ddf=False,
        )
        losses.update(
            {
                k: v * self.weight_dict[k]
                for k, v in local_losses.items()
                if k in self.weight_dict
            }
        )

        num_aux_layers = self.backbone.num_decoder_layers
        auxiliary_outputs_list = [
            {
                "logits": matching_intermediate_logits[:, i, :, :],
                "pred_boxes": keras.ops.clip(
                    matching_intermediate_ref_points[:, i, :, :], 0, 1
                ),
                "pred_corners": matching_predicted_corners[:, i, :, :],
                "ref_points": matching_initial_ref_points[:, i, :, :],
                "teacher_corners": matching_predicted_corners[:, -1, :, :],
                "teacher_logits": matching_intermediate_logits[:, -1, :, :],
            }
            for i in range(num_aux_layers)
        ]
        for i, aux_output in enumerate(auxiliary_outputs_list):
            aux_indices = hungarian_matcher(
                aux_output,
                [targets],
                num_targets_per_image,
                self.use_focal_loss,
                self.matcher_alpha,
                self.matcher_gamma,
                self.matcher_bbox_cost,
                self.matcher_class_cost,
                self.matcher_ciou_cost,
                self.backbone,
            )
            aux_vfl_loss = compute_vfl_loss(
                aux_output,
                [targets],
                aux_indices,
                num_boxes,
                self.num_classes,
                self.matcher_alpha,
                self.matcher_gamma,
            )
            aux_box_losses = compute_box_losses(
                aux_output, [targets], aux_indices, num_boxes
            )
            is_not_last_aux_layer = i < len(auxiliary_outputs_list) - 1
            aux_local_losses = compute_local_losses(
                aux_output,
                [targets],
                aux_indices,
                num_boxes,
                self.backbone,
                self.ddf_temperature,
                compute_ddf=is_not_last_aux_layer,
            )
            aux_losses = {**aux_vfl_loss, **aux_box_losses, **aux_local_losses}
            weighted_aux_losses = {
                k + f"_aux_{i}": v * self.weight_dict[k]
                for k, v in aux_losses.items()
                if k in self.weight_dict
            }
            losses.update(weighted_aux_losses)
        # Add encoder loss.
        enc_output = {
            "logits": enc_topk_logits,
            "pred_boxes": keras.ops.clip(enc_topk_bboxes, 0, 1),
        }
        enc_indices = hungarian_matcher(
            enc_output,
            [targets],
            num_targets_per_image,
            self.use_focal_loss,
            self.matcher_alpha,
            self.matcher_gamma,
            self.matcher_bbox_cost,
            self.matcher_class_cost,
            self.matcher_ciou_cost,
            self.backbone,
        )
        enc_vfl_loss = compute_vfl_loss(
            enc_output,
            [targets],
            enc_indices,
            num_boxes,
            self.num_classes,
            self.matcher_alpha,
            self.matcher_gamma,
        )
        enc_box_losses = compute_box_losses(
            enc_output, [targets], enc_indices, num_boxes
        )
        enc_losses = {**enc_vfl_loss, **enc_box_losses}
        weighted_enc_losses = {
            k + "_enc": v * self.weight_dict[k]
            for k, v in enc_losses.items()
            if k in self.weight_dict
        }
        losses.update(weighted_enc_losses)

        if denoising_meta_values is not None:
            dn_indices = get_cdn_matched_indices(denoising_meta_values)
            dn_num_group = denoising_meta_values["dn_num_group"][0]
            num_boxes_dn = num_boxes * keras.ops.cast(dn_num_group, "float32")
            num_dn_layers = self.backbone.num_decoder_layers + 1
            for i in range(num_dn_layers):
                is_not_last_layer = keras.ops.less(i, num_dn_layers - 1)
                teacher_idx = num_dn_layers - 1
                dn_aux_output = {
                    "logits": dn_intermediate_logits[:, i, :, :],
                    "pred_boxes": keras.ops.clip(
                        dn_intermediate_ref_points[:, i, :, :], 0, 1
                    ),
                    "pred_corners": dn_predicted_corners[:, i, :, :],
                    "ref_points": dn_initial_ref_points[:, i, :, :],
                    "teacher_corners": dn_predicted_corners[
                        :, teacher_idx, :, :
                    ],
                    "teacher_logits": dn_intermediate_logits[
                        :, teacher_idx, :, :
                    ],
                }
                vfl_loss = compute_vfl_loss(
                    dn_aux_output,
                    [targets],
                    dn_indices,
                    num_boxes_dn,
                    self.num_classes,
                    self.matcher_alpha,
                    self.matcher_gamma,
                )
                box_losses = compute_box_losses(
                    dn_aux_output, [targets], dn_indices, num_boxes_dn
                )
                local_losses = compute_local_losses(
                    dn_aux_output,
                    [targets],
                    dn_indices,
                    num_boxes_dn,
                    self.backbone,
                    self.ddf_temperature,
                    compute_ddf=is_not_last_layer,
                )
                all_losses = {**vfl_loss, **box_losses, **local_losses}
                weighted_losses = {
                    k + f"_dn_{i}": v * self.weight_dict[k]
                    for k, v in all_losses.items()
                    if k in self.weight_dict
                }
                losses.update(weighted_losses)
        total_loss = keras.ops.sum([v for v in losses.values()])
        return total_loss

    @property
    def prediction_decoder(self):
        return self._prediction_decoder

    @prediction_decoder.setter
    def prediction_decoder(self, prediction_decoder):
        if prediction_decoder.bounding_box_format != self.bounding_box_format:
            raise ValueError(
                "Expected `prediction_decoder` and `DFineObjectDetector` to "
                "use the same `bounding_box_format`, but got "
                "`prediction_decoder.bounding_box_format="
                f"{prediction_decoder.bounding_box_format}`, and "
                "`self.bounding_box_format="
                f"{self.bounding_box_format}`."
            )
        self._prediction_decoder = prediction_decoder
        self.make_predict_function(force=True)
        self.make_train_function(force=True)
        self.make_test_function(force=True)

    def decode_predictions(self, predictions, data):
        """Decodes raw model predictions into final bounding boxes.

        This method takes the raw output from the model (logits and normalized
        bounding boxes in center format) and converts them into the final
        detection format. The process involves:
        1.  Denormalizing the bounding box coordinates to the original image
            dimensions.
        2.  Converting boxes from center format `(cx, cy, w, h)` to corner
            format `(ymin, xmin, ymax, xmax)`.
        3.  Applying non-maximum suppression to filter out overlapping boxes
            and keep only the most confident detections.

        Args:
            predictions: dict, A dictionary of tensors from the model,
                containing `"logits"` and `"pred_boxes"`.
            data: tuple, The input data tuple, from which the original images
                are extracted to obtain their dimensions for denormalization.

        Returns:
            Dictionary: Final predictions, containing `"boxes"`, `"labels"`,
            `"confidence"`, and `"num_detections"`.
        """
        if isinstance(data, (list, tuple)):
            images, _ = data
        else:
            images = data
        logits = predictions["logits"]
        pred_boxes = predictions["pred_boxes"]
        height, width, _ = keras.ops.shape(images)[1:]
        denormalized_boxes = keras.ops.stack(
            [
                pred_boxes[..., 0] * width,  # center_x
                pred_boxes[..., 1] * height,  # center_y
                pred_boxes[..., 2] * width,  # width
                pred_boxes[..., 3] * height,  # height
            ],
            axis=-1,
        )
        pred_boxes_xyxy = keras.utils.bounding_boxes.convert_format(
            denormalized_boxes,
            source="center_xywh",
            target="xyxy",
        )
        pred_boxes_yxyx = keras.ops.stack(
            [
                pred_boxes_xyxy[..., 1],  # y_min
                pred_boxes_xyxy[..., 0],  # x_min
                pred_boxes_xyxy[..., 3],  # y_max
                pred_boxes_xyxy[..., 2],  # x_max
            ],
            axis=-1,
        )
        y_pred = self.prediction_decoder(pred_boxes_yxyx, logits, images=images)
        return y_pred

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_classes": self.num_classes,
                "bounding_box_format": self.bounding_box_format,
                "matcher_class_cost": self.matcher_class_cost,
                "matcher_bbox_cost": self.matcher_bbox_cost,
                "matcher_ciou_cost": self.matcher_ciou_cost,
                "use_focal_loss": self.use_focal_loss,
                "matcher_alpha": self.matcher_alpha,
                "matcher_gamma": self.matcher_gamma,
                "weight_loss_vfl": self.weight_dict["loss_vfl"],
                "weight_loss_bbox": self.weight_dict["loss_bbox"],
                "weight_loss_ciou": self.weight_dict["loss_ciou"],
                "weight_loss_fgl": self.weight_dict["loss_fgl"],
                "weight_loss_ddf": self.weight_dict["loss_ddf"],
                "ddf_temperature": self.ddf_temperature,
                "prediction_decoder": keras.saving.serialize_keras_object(
                    self._prediction_decoder
                ),
            }
        )
        return config

    def predict_step(self, *args):
        outputs = super().predict_step(*args)
        if isinstance(outputs, tuple):
            return self.decode_predictions(outputs[0], args[-1]), outputs[1]
        return self.decode_predictions(outputs, *args)

    @classmethod
    def from_config(cls, config):
        config = config.copy()
        if "backbone" in config and isinstance(config["backbone"], dict):
            config["backbone"] = keras.saving.deserialize_keras_object(
                config["backbone"]
            )
        if "preprocessor" in config and isinstance(
            config["preprocessor"], dict
        ):
            config["preprocessor"] = keras.saving.deserialize_keras_object(
                config["preprocessor"]
            )
        if "prediction_decoder" in config and isinstance(
            config["prediction_decoder"], dict
        ):
            config["prediction_decoder"] = (
                keras.saving.deserialize_keras_object(
                    config["prediction_decoder"]
                )
            )
        return cls(**config)
